"""Functions for encoding and decoding fixed size integers and floats"""

# Copyright (c) 2018-2024 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import struct
import binascii
import six
from blackboxprotobuf.lib.exceptions import DecoderException, EncoderException

import six

if six.PY3:
    from typing import Any, Tuple


# Generic functions for encoding/decoding structs based on the "struct" format
def encode_struct(fmt, value):
    # type: (str, Any) -> bytes
    """Generic method for encoding arbitrary python "struct" values"""
    try:
        return struct.pack(fmt, value)
    except struct.error as exc:
        six.raise_from(
            EncoderException(
                "Error encoding value %r with format string %s" % (value, fmt)
            ),
            exc,
        )


def decode_struct(fmt, buf, pos):
    # type: (str, bytes, int) -> Tuple[Any, int]
    """Generic method for decoding arbitrary python "struct" values"""
    new_pos = pos + struct.calcsize(fmt)
    try:
        return struct.unpack(fmt, buf[pos:new_pos])[0], new_pos
    except struct.error as exc:
        six.raise_from(
            DecoderException(
                "Error deocding format string %s from bytes: %r"
                % (fmt, binascii.hexlify(buf[pos:new_pos]))
            ),
            exc,
        )


_fixed32_fmt = "<I"

# Note on types: We use Any here for decoded objects. While we could manually
# enforce individual types, we would have to do so via `typing.cast` anyway
# because of `struct`. The Message type also has `Any` anyways.


def encode_fixed32(value):
    # type: (Any) -> bytes
    """Encode a single 32 bit fixed-size value"""
    return encode_struct(_fixed32_fmt, value)


def decode_fixed32(buf, pos):
    # type: (bytes, int) -> Tuple[Any, int]
    """Decode a single 32 bit fixed-size value"""
    return decode_struct(_fixed32_fmt, buf, pos)


_sfixed32_fmt = "<i"


def encode_sfixed32(value):
    # type: (Any) -> bytes
    """Encode a single signed 32 bit fixed-size value"""
    return encode_struct(_sfixed32_fmt, value)


def decode_sfixed32(buf, pos):
    # type: (bytes, int) -> Tuple[Any, int]
    """Decode a single signed 32 bit fixed-size value"""
    return decode_struct(_sfixed32_fmt, buf, pos)


_float_fmt = "<f"


def encode_float(value):
    # type: (Any) -> bytes
    """Encode a single 32 bit floating point value"""
    return encode_struct(_float_fmt, value)


def decode_float(buf, pos):
    # type: (bytes, int) -> Tuple[Any, int]
    """Decode a single 32 bit floating point value"""
    return decode_struct(_float_fmt, buf, pos)


_fixed64_fmt = "<Q"


def encode_fixed64(value):
    # type: (Any) -> bytes
    """Encode a single 64 bit fixed-size value"""
    return encode_struct(_fixed64_fmt, value)


def decode_fixed64(buf, pos):
    # type: (bytes, int) -> Tuple[Any, int]
    """Decode a single 64 bit fixed-size value"""
    return decode_struct(_fixed64_fmt, buf, pos)


_sfixed64_fmt = "<q"


def encode_sfixed64(value):
    # type: (Any) -> bytes
    """Encode a single signed 64 bit fixed-size value"""
    return encode_struct(_sfixed64_fmt, value)


def decode_sfixed64(buf, pos):
    # type: (bytes, int) -> Tuple[Any, int]
    """Decode a single signed 64 bit fixed-size value"""
    return decode_struct(_sfixed64_fmt, buf, pos)


_double_fmt = "<d"


def encode_double(value):
    # type: (Any) -> bytes
    """Encode a single 64 bit floating point value"""
    return encode_struct(_double_fmt, value)


def decode_double(buf, pos):
    # type: (bytes, int) -> Tuple[Any, int]
    """Decode a single 64 bit floating point value"""
    return decode_struct(_double_fmt, buf, pos)



"""Tests similar to the length_delim or protobuf tests, but make sure we can round trip through the JSON encode/decode """

# Copyright (c) 2018-2024 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from hypothesis import given, assume, note, example, reproduce_failure
import hypothesis.strategies as st
import strategies
import json
import six
import binascii

from blackboxprotobuf.lib.config import Config
from blackboxprotobuf.lib.types import length_delim
from blackboxprotobuf.lib.types import type_maps
from blackboxprotobuf.lib.typedef import TypeDef
from blackboxprotobuf.lib.payloads import grpc, gzip
import blackboxprotobuf


@given(x=strategies.gen_message())
def test_message_json_inverse(x):
    config = Config()
    typedef, message = x
    encoded = length_delim.encode_message(message, config, TypeDef.from_dict(typedef))
    decoded_json, typedef_out = blackboxprotobuf.protobuf_to_json(
        encoded, config=config, message_type=typedef
    )
    blackboxprotobuf.validate_typedef(typedef_out)
    encoded_json = blackboxprotobuf.protobuf_from_json(
        decoded_json, config=config, message_type=typedef_out
    )
    assert not isinstance(encoded_json, list)
    decoded, typedef_out = blackboxprotobuf.decode_message(
        encoded_json, config=config, message_type=typedef
    )
    blackboxprotobuf.validate_typedef(typedef_out)
    assert isinstance(encoded, bytearray)
    assert isinstance(decoded, dict)
    assert message == decoded


@given(x=strategies.gen_message(), n=st.integers(min_value=2, max_value=10))
def test_multiple_encoding(x, n):
    config = Config()
    typedef, message = x
    encoded = length_delim.encode_message(message, config, TypeDef.from_dict(typedef))

    bufs = [encoded] * n
    message_json, typedef_out = blackboxprotobuf.protobuf_to_json(bufs, typedef, config)
    messages = json.loads(message_json)
    assert isinstance(messages, list)
    assert len(messages) == n

    encoded2 = blackboxprotobuf.protobuf_from_json(message_json, typedef, config)
    assert isinstance(encoded2, list)
    assert len(encoded2) == n


@given(x=strategies.gen_message(anon=True))
def test_anon_json_decode(x):
    """Create a new encoded message, the try to decode without a typedef into a
    json, from json back to binary and then  finally decode the message back to
    the original format. Makes sure json decoding can handle any types and does
    not change the essage.
    """
    config = Config()
    typedef, message = x
    encoded = blackboxprotobuf.encode_message(
        message, config=config, message_type=typedef
    )
    decoded_json, typedef_out = blackboxprotobuf.protobuf_to_json(
        encoded, config=config
    )
    blackboxprotobuf.validate_typedef(typedef_out)
    note("To Json Typedef: %r" % dict(typedef_out))
    encoded_json = blackboxprotobuf.protobuf_from_json(
        decoded_json, config=config, message_type=typedef_out
    )
    assert not isinstance(encoded_json, list)
    decoded, typedef_out = blackboxprotobuf.decode_message(
        encoded_json, config=config, message_type=typedef
    )
    blackboxprotobuf.validate_typedef(typedef_out)
    note("Original message: %r" % message)
    note("Decoded JSON: %r" % decoded_json)
    note("Decoded message: %r" % decoded)
    note("Original typedef: %r" % typedef)
    note("Decoded typedef: %r" % typedef_out)

    def check_message(orig, orig_typedef, new, new_typedef):
        for field_number in set(orig.keys()) | set(new.keys()):
            # verify all fields are there
            assert field_number in orig
            assert field_number in orig_typedef
            assert field_number in new
            assert field_number in new_typedef

            orig_values = orig[field_number]
            new_values = new[field_number]
            orig_type = orig_typedef[field_number]["type"]
            new_type = new_typedef[field_number]["type"]

            note("Parsing field# %s" % field_number)
            note("orig_values: %r" % orig_values)
            note("new_values: %r" % new_values)
            note("orig_type: %s" % orig_type)
            note("new_type: %s" % new_type)
            # Fields might be lists. Just convert everything to a list
            if not isinstance(orig_values, list):
                orig_values = [orig_values]
                assert not isinstance(new_values, list)
                new_values = [new_values]
            assert isinstance(orig_values, list)
            assert isinstance(new_values, list)

            # if the types don't match, then try to convert them
            if new_type == "message" and orig_type in ["bytes", "string"]:
                # if the type is a message, we want to convert the orig type to a message
                # this isn't ideal, we'll be using the unintended type, but
                # best way to compare. Re-encoding a  message to binary might
                # not keep the field order
                new_field_typedef = new_typedef[field_number]["message_typedef"]
                for i, orig_value in enumerate(orig_values):
                    if orig_type == "bytes":
                        (
                            orig_values[i],
                            orig_field_typedef,
                            _,
                            _,
                        ) = length_delim.decode_lendelim_message(
                            length_delim.encode_bytes(orig_value),
                            config,
                            TypeDef.from_dict(new_field_typedef),
                        )
                        orig_field_typedef = orig_field_typedef.to_dict()
                    else:
                        # string value
                        (
                            orig_values[i],
                            orig_field_typedef,
                            _,
                            _,
                        ) = length_delim.decode_lendelim_message(
                            length_delim.encode_string(orig_value),
                            config,
                            TypeDef.from_dict(new_field_typedef),
                        )
                        orig_field_typedef = orig_field_typedef.to_dict()
                    orig_typedef[field_number]["message_typedef"] = orig_field_typedef
                orig_type = "message"

            if new_type == "string" and orig_type == "bytes":
                # our bytes were accidently valid string
                new_type = "bytes"
                for i, new_value in enumerate(new_values):
                    new_values[i], _ = length_delim.decode_bytes(
                        length_delim.encode_string(new_value), 0
                    )
            note("New values: %r" % new_values)
            # sort the lists with special handling for dicts
            orig_values.sort(key=lambda x: x if not isinstance(x, dict) else x.items())
            new_values.sort(key=lambda x: x if not isinstance(x, dict) else x.items())
            for orig_value, new_value in zip(orig_values, new_values):
                if orig_type == "message":
                    check_message(
                        orig_value,
                        orig_typedef[field_number]["message_typedef"],
                        new_value,
                        new_typedef[field_number]["message_typedef"],
                    )
                else:
                    assert orig_value == new_value

    check_message(message, typedef, decoded, typedef_out)
    # assert message == decoded



from burp import IMessageEditorTab
from javax.swing import JTabbedPane, JPanel, JCheckBox, JButton
from java.awt import GridLayout, BorderLayout
from java.awt.event import ActionListener
import traceback

import grpc_coder
from grpc_coder_withdout_dependency import decode_b64_grpc_web_text, encode_grpc_web_json_to_b64_format
from grpc_coder_withdout_dependency import decode_grpc_web_proto_payload, encode_grpc_web_proto_json_to_proto_format
from grpc_coder_withdout_dependency import create_bbpb_type_def_from_json, get_main_json_from_type_def_ordered_dict

class GrpcWebExtensionEditorTab(IMessageEditorTab, ActionListener):  # FIXED: Implement ActionListener
    def __init__(self, extender, controller, editable):
        self._extender = extender
        self._editable = editable
        self._is_first_time_tab_opened = True

        ## main (general) settings checkboxes from extender class to access the easier
        self._enable_extension_msg_editor_tab_check_box = self._extender._enable_extension_msg_editor_tab_check_box
        self._detect_encode_decode_format_from_ct_header_check_box = self._extender._detect_encode_decode_format_from_ct_header_check_box
        self._detect_encode_decode_format_from_x_grpc_header_check_box = self._extender._detect_encode_decode_format_from_x_grpc_header_check_box
        self._enable_application_grpc_web_text_decode_encode_by_default_check_box = self._extender._enable_application_grpc_web_text_decode_encode_by_default_check_box
        self._enable_application_grpc_web_proto_decode_encode_by_default_check_box = self._extender._enable_application_grpc_web_proto_decode_encode_by_default_check_box
        self._enableGrpcWebTextEncodeDecode = False
        self._enableGrpcWebProtoEncodeDecode = False

        self._tabbedPane = JTabbedPane()

        # Payload Tab
        self._txtInputPayload = extender._callbacks.createTextEditor()
        self._txtInputPayload.setEditable(editable)
        self._tabbedPane.addTab("Payload", self._txtInputPayload.getComponent())

        # Type Definition Tab
        self._txtInputTypeDef = extender._callbacks.createTextEditor()
        self._txtInputTypeDef.setEditable(False)

        self._typeDefPanel = JPanel()
        self._typeDefPanel.setLayout(BorderLayout())

        # Buttons for Type Definition
        self._editButton = JButton("Edit Type Definition", actionPerformed=self.actionPerformed)
        self._saveButton = JButton("Save Type Definition", actionPerformed=self.actionPerformed)
        self._resetButton = JButton("Reset Type Definition (Auto)", actionPerformed=self.actionPerformed)
        self._saveButton.setEnabled(False)
        self._isTypeDefinitionEdited = False

        # Add editor to the Type Definition panel
        self._typeDefPanel.add(self._txtInputTypeDef.getComponent(), BorderLayout.CENTER)

        # Create button panel
        self._buttonPanel = JPanel()
        self._buttonPanel.add(self._editButton)
        self._buttonPanel.add(self._saveButton)
        self._buttonPanel.add(self._resetButton)

        # Add button panel to the bottom of Type Definition tab
        self._typeDefPanel.add(self._buttonPanel, BorderLayout.SOUTH)
        self._tabbedPane.addTab("Type Definition", self._typeDefPanel)

        # Settings Tab
        # self._txtInputSettings = JPanel()
        # self._txtInputSettings.setLayout(GridLayout(1, 1))
        # self._grpcWebTextPayloadCheckBox = JCheckBox("application/grpc-web-text payload ?")
        # self._txtInputSettings.add(self._grpcWebTextPayloadCheckBox)
        # self._tabbedPane.addTab("Settings", self._txtInputSettings)

    def print_output(self, text):
        self._extender.print_output(text)

    def print_error(self, text):
        self._extender.print_error(text)

    def getTabCaption(self):
        return "Decoded gRPC-Web ProtoBuf"

    def getUiComponent(self):
        return self._tabbedPane

    def isEnabled(self, content, isRequest):
        if self._enable_application_grpc_web_text_decode_encode_by_default_check_box.isSelected():
            self._enableGrpcWebProtoEncodeDecode = False
            self._enableGrpcWebTextEncodeDecode = True
            return True

        if self._enable_application_grpc_web_proto_decode_encode_by_default_check_box.isSelected():
            self._enableGrpcWebTextEncodeDecode = False
            self._enableGrpcWebProtoEncodeDecode = True
            return True

        __isEnabled = False
        analyzed_request = self._extender._helpers.analyzeRequest(content)
        req_headers = analyzed_request.getHeaders()

        if self._enable_extension_msg_editor_tab_check_box.isSelected():
            __isEnabled = True

        if (self._detect_encode_decode_format_from_ct_header_check_box.isSelected() or
                self._detect_encode_decode_format_from_x_grpc_header_check_box.isSelected()):
            for h in req_headers:
                if self._detect_encode_decode_format_from_ct_header_check_box.isSelected():
                    if h.lower().startswith('content-type'):
                        _, value = h.split(':', 1)
                        value = value.strip()
                        if value.startswith('application/grpc-web-text'):
                            self._enableGrpcWebTextEncodeDecode = True
                            __isEnabled = True
                        elif value.startswith('application/grpc-web+proto'):
                            self._enableGrpcWebProtoEncodeDecode = True
                            __isEnabled = True

                elif self._detect_encode_decode_format_from_x_grpc_header_check_box.isSelected():
                    if h.lower().startswith('x-grpc-content-type'):
                        _, value2 = h.split(':', 1)
                        value2 = value2.strip()
                        if value2.startswith('application/grpc-web-text'):
                            self._enableGrpcWebTextEncodeDecode = True
                            __isEnabled = True
                        elif value2.startswith('application/grpc-web+proto'):
                            self._enableGrpcWebProtoEncodeDecode = True
                            __isEnabled = True

        return __isEnabled

    def isModified(self):
        """ Check if either tab content is modified """
        return self._txtInputPayload.isTextModified() or self._txtInputTypeDef.isTextModified()

    def isGrpcWebTextPayloadEnabled(self):
        """ Returns whether the checkbox is checked """
        return self._grpcWebTextPayloadCheckBox.isSelected()

    def setMessage(self, content, isRequest):
        if content is None:
            self._txtInputPayload.setText(None)
            self._txtInputPayload.setEditable(False)
            self._txtInputTypeDef.setText(None)
            self._txtInputTypeDef.setEditable(False)
            return

        if self._enableGrpcWebTextEncodeDecode:
            analyzed_request = self._extender._helpers.analyzeRequest(content)
            body_offset = analyzed_request.getBodyOffset()
            request_body = content[body_offset:]

            try:
                message, typedef = decode_b64_grpc_web_text(payload=request_body)
                decoded_string = message.decode("unicode_escape")
                message = decoded_string
                message = message.encode('utf-8')
                typedef_main_json = get_main_json_from_type_def_ordered_dict(type_def=typedef)
            except Exception as e:
                message = "Error decoding request: {}".format(str(e))
                typedef_main_json = "No Type Definition"
                self.print_error(message)

            self._txtInputPayload.setText(message)
            self._txtInputPayload.setEditable(self._editable)

            if not self._isTypeDefinitionEdited:
                self._txtInputTypeDef.setText(str(typedef_main_json))
                self._txtInputTypeDef.setEditable(False)

        elif self._enableGrpcWebProtoEncodeDecode:
            analyzed_request = self._extender._helpers.analyzeRequest(content)
            body_offset = analyzed_request.getBodyOffset()
            request_body = content[body_offset:]
            request_body = request_body.tostring()
            request_body = request_body.decode('utf-8', errors='surrogatepass')
            # request_body = request_body.decode("unicode_escape")

            try:
                message, typedef = decode_grpc_web_proto_payload(payload=request_body)
                decoded_string = message.decode("unicode_escape")
                message = decoded_string
                message = message.encode('utf-8')
                typedef_main_json = get_main_json_from_type_def_ordered_dict(type_def=typedef)
            except Exception as e:
                message = "Error decoding request: {}".format(str(e))
                typedef_main_json = "No Type Definition"
                self.print_error(message)

            self._txtInputPayload.setText(message)
            self._txtInputPayload.setEditable(self._editable)

            if not self._isTypeDefinitionEdited:
                self._txtInputTypeDef.setText(str(typedef_main_json))
                self._txtInputTypeDef.setEditable(False)

        self._currentMessage = content

    def actionPerformed(self, event):
        """ Handle button clicks """
        source = event.getSource()
        if source == self._editButton:
            self._txtInputTypeDef.setEditable(True)
            self._saveButton.setEnabled(True)
            self.print_output("[*] Edit mode enabled")

        elif source == self._saveButton:
            content = self._txtInputTypeDef.getText()
            self.print_output("[*] Type Definition Saved")
            self._txtInputTypeDef.setEditable(False)
            self._saveButton.setEnabled(False)
            self._isTypeDefinitionEdited = True

        elif source == self._resetButton:
            self._isTypeDefinitionEdited = False
            content = self._txtInputTypeDef.getText()
            self._txtInputTypeDef.setEditable(False)
            self._saveButton.setEnabled(False)
            self.print_output("[*] Type Definition is Reset")

    def getMessage(self):
        """ Return the modified content from the payload tab """
        if self._enableGrpcWebTextEncodeDecode:
            try:
                modified_payload = self._txtInputPayload.getText()  # Get modified text
                modified_payload = modified_payload.tostring()
                # modified_payload = modified_payload.decode('utf-8')
                type_def_raw = self._txtInputTypeDef.getText().tostring().decode('utf-8')
                type_def_object = create_bbpb_type_def_from_json(type_def_raw)
                encoded_payload = encode_grpc_web_json_to_b64_format(modified_payload, type_def_object)  # Convert back

                # Get original request headers
                original_request = self._extender._helpers.analyzeRequest(self._currentMessage)
                headers = original_request.getHeaders()

                # Construct the new request
                new_request = self._extender._helpers.buildHttpMessage(headers, encoded_payload)
                return new_request

            except Exception as e:
                self.print_error("Error encoding modified payload:" +  str(e))

        elif self._enableGrpcWebProtoEncodeDecode:
            try:
                modified_payload = self._txtInputPayload.getText()  # Get modified text
                modified_payload = modified_payload.tostring()
                # modified_payload = modified_payload.decode('utf-8')
                type_def_raw = self._txtInputTypeDef.getText().tostring().decode('utf-8')
                type_def_object = create_bbpb_type_def_from_json(type_def_raw)
                encoded_payload = encode_grpc_web_proto_json_to_proto_format(modified_payload, type_def_object)  # Convert back
                encoded_payload = bytes(encoded_payload)

                # Get original request headers
                original_request = self._extender._helpers.analyzeRequest(self._currentMessage)
                headers = original_request.getHeaders()

                # Construct the new request
                new_request = self._extender._helpers.buildHttpMessage(headers, encoded_payload)
                return new_request

            except Exception as e:
                self.print_error("Error encoding modified payload:" +  str(e))

        # Return the original request if no modifications
        return self._currentMessage



# Copyright (c) 2018-2024 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import six
import struct
from blackboxprotobuf.lib.exceptions import BlackboxProtobufException

if six.PY3:
    from typing import Tuple

# gRPC over HTTP2 spec: https://github.com/grpc/grpc/blob/master/doc/PROTOCOL-HTTP2.md

HEADER_LEN = 1 + 4


def is_grpc(payload):
    # type: (bytes) -> bool
    if len(payload) < HEADER_LEN:
        return False
    if six.PY2 and isinstance(payload, bytearray):
        payload = bytes(payload)
    pos = 0
    while pos < len(payload):
        compression_byte = six.indexbytes(payload, pos)
        # Change this to support 0x1 once we support compression
        if compression_byte != 0:
            return False
        message_length = struct.unpack_from(">I", payload[pos + 1 : pos + 5])[0]
        pos += message_length + 5

    if pos != len(payload):
        return False
    return True


def decode_grpc(payload):
    # type: (bytes) -> Tuple[bytes | list[bytes], str]
    """Decode GRPC. Return the protobuf data"""
    if six.PY2 and isinstance(payload, bytearray):
        payload = bytes(payload)

    if len(payload) == 0:
        raise BlackboxProtobufException("Error decoding GRPC. Payload is empty")

    pos = 0
    payloads = []
    while pos + HEADER_LEN <= len(payload):
        compression_byte = six.indexbytes(payload, pos)
        pos += 1
        if compression_byte != 0x00:
            if compression_byte == 0x01:
                # Payload is compressed
                # If a payload is compressed, the compression method is specified in the `grpc-encoding` header
                # Options are  "identity" / "gzip" / "deflate" / "snappy" / {custom}
                raise BlackboxProtobufException(
                    "Error decoding GRPC. Compressed payloads are not supported"
                )
            else:
                raise BlackboxProtobufException(
                    "Error decoding GRPC. First byte must be 0 or 1 to indicate compression"
                )

        message_length = struct.unpack_from(">I", payload[pos : pos + 4])[0]
        pos += 4

        if len(payload) < pos + message_length:
            raise BlackboxProtobufException(
                "Error decoding GRPC. Payload length does not match encoded gRPC length"
            )

        payloads.append(payload[pos : pos + message_length])
        pos += message_length

    if pos != len(payload):
        raise BlackboxProtobufException(
            "Error decoding GRPC. Payload length does not match encoded gRPC lengths"
        )

    if len(payloads) > 1:
        return payloads, "grpc"
    else:
        return payloads[0], "grpc"


def encode_grpc(data, encoding="grpc"):
    # type: (bytes | list[bytes], str) -> bytes
    if encoding != "grpc":
        raise BlackboxProtobufException(
            "Error encoding GRPC with encoding %s. GRPC is only supported with no compression"
            % encoding
        )

    datas = data if isinstance(data, list) else [data]

    payload = bytearray()
    for data in datas:
        payload.append(0x00)  # No compression
        payload.extend(struct.pack(">I", len(data)))  # Length
        payload.extend(data)

    return payload



# Copyright (c) 2018-2024 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .types import type_maps

import six
from blackboxprotobuf.lib.exceptions import (
    DecoderException,
)

if six.PY3:
    import typing

    if typing.TYPE_CHECKING:
        from typing import Dict
        from .pytypes import TypeDefDict


class Config:
    def __init__(self):
        # type: (Config) -> None
        # Map of message type names to typedefs, previously stored at
        # `blackboxprotobuf.known_messages`
        self.known_types = {}  # type: Dict[str, TypeDefDict]

        # Default type for "bytes" like objects that aren't messages or strings
        # Other option is currently just 'bytes_hex'
        self.default_binary_type = "bytes"

        # Change the default type for a wiretype (eg. change ints to be signed
        # by default or fixed fields to always be float)
        self.default_types = {}  # type: Dict[int, str]

        # Configure whether bbpb should try to re-encode fields in the same
        # order they decoded
        # Field order shouldn't matter for real protobufs, but is there to ensure
        # that bytes/string are accidentally valid protobufs don't get scrambled
        # by decoding/re-encoding
        self.preserve_field_order = True

    def get_default_type(self, wiretype):
        # type: (Config, int) -> str
        default_type = self.default_types.get(wiretype, None)
        if default_type is None:
            default_type = type_maps.WIRE_TYPE_DEFAULTS.get(wiretype, None)

        if default_type is None:
            raise DecoderException(
                "Could not find default type for wire type %d" % wiretype
            )
        return default_type


default = Config()



"""The `blackboxprotobuf.lib.api` module provides high level functions for
decoding and re-encoding protobuf messages.

Most functions take the input data, a type definition and a config object.

The 'message_type' or type definition (typedef) is a blackboxprotobuf specific format
which defines which types should be used for decoding/encoding each field. It
is optional for decoding functions but required for encoding funtions. The
decoding function will return a typedef that is require to re-encode the array.
If a typedef was provided during decoding, then those types will be used for
decoding and the typedef return will be the original typedef + any new fields
in the message.

The config argument is the Config object from `blackboxprotobuf.lib.config` and
allows reconfiguring default types and stores "known" message typedefs that can
be referenced within other typedefs. This argument can be left out to use a
default shared config object.
"""

# Copyright (c) 2018-2024 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import six
import json
import collections
import blackboxprotobuf.lib.protofile
import blackboxprotobuf.lib.types.length_delim
import blackboxprotobuf.lib.types.type_maps
from blackboxprotobuf.lib.config import default as default_config
from blackboxprotobuf.lib.exceptions import (
    TypedefException,
    EncoderException,
    DecoderException,
)
from blackboxprotobuf.lib.typedef import TypeDef

if six.PY3:
    import typing

    # Circular imports on Config if we don't check here
    if typing.TYPE_CHECKING:
        from typing import Dict, List, Optional
        from blackboxprotobuf.lib.pytypes import Message, TypeDefDict, FieldDefDict
        from blackboxprotobuf.lib.config import Config


def decode_message(buf, message_type=None, config=None):
    # type: (bytes, Optional[str | TypeDefDict], Optional[Config]) -> tuple[Message, TypeDefDict]
    """Decode a protobuf message and return a python dictionary representing
    the message.

    Args:
        buf: Bytes representing an encoded protobuf message
        message_type: Optional type to use as the base for decoding. Allows for
            customizing field types or names. Can be a python dictionary or a
            message type name which maps to the `known_types` dictionary in the
            config. Defaults to an empty definition '{}'.
        config: `blackboxprotobuf.lib.config.Config` object which allows
            customizing default types for wire types and holds the
            `known_types` array. Defaults to
            `blackboxprotobuf.lib.config.default` if not provided.
    Returns:
        A tuple containing a python dictionary representing the message and a
        type definition for re-encoding the message.

        The type definition is based on the `message_type` argument if one was
        provided, but may add additional fields if new fields were encountered
        during decoding.
    """

    if config is None:
        config = default_config

    if isinstance(buf, bytearray):
        buf = bytes(buf)
    buf = six.ensure_binary(buf)
    if message_type is None:
        message_type = {}
    elif isinstance(message_type, six.string_types):
        if message_type not in config.known_types:
            message_type = {}
        else:
            message_type = config.known_types[message_type]

    if not isinstance(message_type, dict):
        raise DecoderException(
            "Decode message received an invalid typedef type. Typedef should be a string with a message name, a dictionary, or None"
        )
    value, typedef, _, _ = blackboxprotobuf.lib.types.length_delim.decode_message(
        buf, config, TypeDef.from_dict(message_type)
    )
    return value, typedef.to_dict()


def encode_message(value, message_type, config=None):
    # type: (Message, str | TypeDefDict, Optional[Config]) -> bytes
    """Re-encode a python dictionary as a binary protobuf message.

    Args:
        value: Python dictionary to re-encode to bytes. This should usually be
            a modified version of the dictionary returned by `decode_message`.
        message_type: Type definition to use to re-encode the message. This
            will should generally be the type definition returned from the
            original `decode_message` call.
        config: `blackboxprotobuf.lib.config.Config` object which allows
            customizing default types for wire types and holds the
            `known_types` array. Defaults to
            `blackboxprotobuf.lib.config.default` if not provided.
    Returns:
        A bytearray containing the encoded protobuf message.
    """

    if config is None:
        config = default_config

    if message_type is None:
        raise EncoderException(
            "Encode message must have valid type definition. message_type cannot be None"
        )

    if isinstance(message_type, six.string_types):
        if message_type not in config.known_types:
            raise EncoderException(
                "The provided message type name (%s) is not known. Encoding requires a valid type definition"
                % message_type
            )
        message_type = config.known_types[message_type]

    if not isinstance(message_type, dict):
        raise EncoderException(
            "Encode message received an invalid typedef type. Typedef should be a string with a message name or a dictionary."
        )
    return bytes(
        blackboxprotobuf.lib.types.length_delim.encode_message(
            value, config, TypeDef.from_dict(message_type)
        )
    )


def protobuf_to_json(buf, message_type=None, config=None):
    # type: (bytes | list[bytes], Optional[str | TypeDefDict], Optional[Config]) -> tuple[str, TypeDefDict]
    """Decode a protobuf messages and return a JSON string representing the
    messages.

    Args:
        buf: One or more bytes representing encoded protobuf messages
        message_type: Optional type to use as the base for decoding. Allows for
            customizing field types or names. Can be a python dictionary or a
            message type name which maps to the `known_types` dictionary in the
            config. Defaults to an empty definition '{}'.
        config: `blackboxprotobuf.lib.config.Config` object which allows
            customizing default types for wire types and holds the
            `known_types` array. Defaults to
            `blackboxprotobuf.lib.config.default` if not provided.
    Returns:
        A tuple containing a JSON string representing the messages and a type
        definition for re-encoding the messages.

        The JSON string and type definition are annotated and sorted for
        readability.

        The type definition is based on the `message_type` argument if one was
        provided, but may add additional fields if new fields were encountered
        during decoding.
    """
    values = []
    bufs = buf if isinstance(buf, list) else [buf]

    if len(bufs) == 0:
        raise DecoderException("No protobuf bytes were provided")

    for data in bufs:
        value, message_type = decode_message(data, message_type, config)
        value = _json_safe_transform(value, message_type, False, config=config)
        value = _sort_output(value, message_type, config=config)
        values.append(value)

    if not isinstance(message_type, dict):
        # Shouldn't happen because of len(bufs) check, but make the type checker happy and verify edge cases
        raise DecoderException(
            "Error decoding to json: Could not find valid message_type type (dict). Found: %s"
            % type(message_type)
        )
    _annotate_typedef(message_type, values[0])
    message_type = sort_typedef(message_type)

    if not isinstance(buf, list) and len(values) == 1:
        return json.dumps(values[0], indent=2), message_type
    else:
        return json.dumps(values, indent=2), message_type


def protobuf_from_json(json_str, message_type, config=None):
    # type: (str, str | TypeDefDict, Optional[Config]) -> bytes | list[bytes]
    """Re-encode a JSON string as a binary protobuf message.

    Args:
        json_str: JSON string to re-encode to protobuf message bytes. This
            should usually be a modified version of the value returned by
            `protobuf_to_json`.
        message_type: Type definition to use to re-encode the message. This
            will should generally be the type definition returned from the
            original `protobuf_to_json` call.
        config: `blackboxprotobuf.lib.config.Config` object which allows
            customizing default types for wire types and holds the
            `known_types` array. Defaults to
            `blackboxprotobuf.lib.config.default` if not provided.
    Returns:
        A bytearray containing the encoded protobuf message.
    """
    if config is None:
        config = default_config
    if isinstance(message_type, six.string_types):
        if message_type not in config.known_types:
            raise EncoderException(
                'protobuf_from_json must have valid type definition. message_type "%s" is not known'
                % message_type
            )
        message_type = config.known_types[message_type]
    if not isinstance(message_type, dict):
        raise EncoderException(
            "Encode message received an invalid typedef type. Typedef should be a string with a message name or a dictionary."
        )

    value = json.loads(json_str)
    values = value if isinstance(value, list) else [value]

    _strip_typedef_annotations(message_type)
    values = [_json_safe_transform(message, message_type, True) for message in values]

    payloads = []
    for message in values:
        payloads.append(encode_message(message, message_type, config))

    if not isinstance(value, list) and len(payloads) == 1:
        return payloads[0]
    else:
        return payloads


def export_protofile(message_types, output_filename):
    # type: (Dict[str, TypeDefDict], str) -> None
    """This function attempts to export a set of message type definitions to a
    `.proto` file for use with other tools.

    Args:
        message_types: Python dictionary containing the type definitions to
            export. The dictionary should contain the message type name as the
            key and the type definition as the value.
        output_filename: String representing the filename to output the
            protobuf definition file to.
    """
    blackboxprotobuf.lib.protofile.export_proto(
        message_types, output_filename=output_filename
    )


def import_protofile(input_filename, save_to_known=True, config=None):
    # type: (str, bool, Optional[Config]) -> Dict[str, TypeDefDict] | None
    """This function attempts to import a set of message type definitions from a
    `.proto` file.

    This is a convenience function for `blackboxprotobuf.lib.protofile`. The
    protobuf file import support is not complete and may fail for some type
    definitions.

    Args:
        input_filename: Filename to read the protobuf definitions from.
        save_to_known: If True, this function will save the message type
            definitions to `config.known_types`. Otherwise, it will return them
            to the caller. Defaults to `True`.
        config: Optional config object which stores the `known_types` map.
            Defaults to `blackboxprotobuf.lib.config.default`.
    Returns:
        If `save_to_known` is False, then the type definitions read from the
        file are returned as a dictionary, with the type names as the keys and
        type definitions as the values.
    """
    if config is None:
        config = default_config

    new_typedefs = blackboxprotobuf.lib.protofile.import_proto(
        config, input_filename=input_filename
    )
    if save_to_known:
        config.known_types.update(new_typedefs)
        return None
    else:
        return new_typedefs


NAME_REGEX = re.compile(r"\A[a-zA-Z][a-zA-Z0-9_]*\Z")


def validate_typedef(typedef, old_typedef=None, config=None, _path=None):
    # type: (TypeDefDict, Optional[TypeDefDict], Optional[Config], Optional[List[str]]) -> None
    """Attempt to validate a type definition object is valid.

    This function attempts to ensure a type definition is valid before it is
    used to encode/decode a message. This will make sure the field names are
    valid and field names/numbers are consistent. It is intended to be called
    after a user has edited the type definition to ensure the edits are valid.

    Args:
        typedef: The type definition object to validate. This should be a
            python dict derived from the dict returned  by a decode function.
        old_typedef: Optionally provide a old version of the type definition to
            compare the new type definnition to. If provided, this function
            will ensure any type changes are valid. For example, a field with a
            varint type can be changed to other varint types, but not a string
            or float.
        config: Optionally provide a config object which contains the
            `known_types` map to map message type names to known type definitions.
            Defaults to `blackboxprotobuf.lib.config.default`.
    Raises:
        TypedefException: Raises a TypedefException if the provided type
            definition is not valid.
    """
    if _path is None:
        _path = []
    if config is None:
        config = default_config

    int_keys = set()
    field_names = set()
    for field_number, field_typedef in typedef.items():
        alt_field_number = None
        if "-" in str(field_number):
            field_number, alt_field_number = field_number.split("-")

        # Validate field_number is a number
        if not str(field_number).isdigit():
            raise TypedefException("Field number must be a digit: %s" % field_number)
        field_number = six.ensure_text(str(field_number))

        field_path = _path[:]
        field_path.append(field_number)

        # Check for duplicate field numbers
        if field_number in int_keys:
            raise TypedefException(
                "Duplicate field number: %s" % field_number, field_path
            )
        int_keys.add(field_number)

        # Must have a type field
        if "type" not in field_typedef:
            raise TypedefException(
                "Field number must have a type value: %s" % field_number, field_path
            )
        if alt_field_number is not None:
            if field_typedef["type"] != "message":
                raise TypedefException(
                    "Alt field number (%s) specified for non-message field: %s"
                    % (alt_field_number, field_number),
                    field_path,
                )

        valid_type_fields = [
            "type",
            "name",
            "field_order",
            "message_typedef",
            "message_type_name",
            "alt_typedefs",
            "example_value_ignored",
            "seen_repeated",
        ]
        for key, value in field_typedef.items():
            # Check field keys against valid values
            if key not in valid_type_fields:
                raise TypedefException(
                    'Invalid field key "%s" for field number %s' % (key, field_number),
                    field_path,
                )
            if (
                key in ["message_typedef", "message_type_name"]
                and not field_typedef["type"] == "message"
            ):
                raise TypedefException(
                    'Invalid field key "%s" for field number %s' % (key, field_number),
                    field_path,
                )
            if key == "group_typedef" and not field_typedef["type"] == "group":
                raise TypedefException(
                    'Invalid field key "%s" for field number %s' % (key, field_number),
                    field_path,
                )

            # Validate type value
            if key == "type":
                if value not in blackboxprotobuf.lib.types.type_maps.WIRETYPES:
                    raise TypedefException(
                        'Invalid type "%s" for field number %s' % (value, field_number),
                        field_path,
                    )
            # Check for duplicate names
            if key == "name":
                if not isinstance(value, six.string_types):
                    raise TypedefException(
                        "Invalid type for name field in typedef: %r. Field number %s"
                        % (value, field_number),
                        field_path,
                    )
                if value.strip() == "":
                    continue

                if value.lower() in field_names:
                    raise TypedefException(
                        ('Duplicate field name "%s" for field ' "number %s")
                        % (value, field_number),
                        field_path,
                    )
                if not NAME_REGEX.match(value):
                    raise TypedefException(
                        (
                            'Invalid field name "%s" for field '
                            "number %s. Should match %s"
                        )
                        % (value, field_number, "[a-zA-Z_][a-zA-Z0-9_]*"),
                        field_path,
                    )
                field_names.add(value.lower())

            # Check if message type name is known
            if key == "message_type_name":
                if value not in config.known_types:
                    raise TypedefException(
                        (
                            'Message type "%s" for field number'
                            " %s is not known. Known types: %s"
                        )
                        % (value, field_number, config.known_types.keys()),
                        field_path,
                    )

            # Recursively validate inner typedefs
            if key == "message_typedef":
                if isinstance(value, dict):
                    if (
                        old_typedef is not None
                        and field_number in old_typedef
                        and key in old_typedef[field_number]
                    ):
                        validate_typedef(
                            value,
                            old_typedef=old_typedef[field_number]["message_typedef"],
                            _path=field_path,
                            config=config,
                        )
                    else:
                        validate_typedef(value, _path=field_path, config=config)
            if key == "alt_typedefs":
                for alt_field_number, alt_typedef in field_typedef[
                    "alt_typedefs"
                ].items():
                    if isinstance(alt_typedef, dict):
                        validate_typedef(alt_typedef, _path=field_path, config=config)

    if old_typedef is not None:
        wiretype_map = {}
        for field_number, value in old_typedef.items():
            wiretype_map[
                int(field_number)
            ] = blackboxprotobuf.lib.types.type_maps.WIRETYPES[value["type"]]
        for field_number, value in typedef.items():
            field_path = _path[:]
            field_path.append(str(field_number))
            if int(field_number) in wiretype_map:
                old_wiretype = wiretype_map[int(field_number)]
                if (
                    old_wiretype
                    != blackboxprotobuf.lib.types.type_maps.WIRETYPES[value["type"]]
                ):
                    raise TypedefException(
                        (
                            "Wiretype for field number %s does"
                            " not match old type definition"
                        )
                        % field_number,
                        field_path,
                    )


def _json_safe_transform(values, typedef, toBytes, config=None):
    # type: (Message, TypeDefDict, bool, Optional[Config]) -> Message
    # Python's JSON doesn't have a default way to handle 'bytes' types. To
    # handle this, we want some string like encoding which JSON can handle but
    # can also handle arbitrary bytes. This method get's more complicated than
    # just converting all bytes since on re-encoding we need to know which ones
    # were transformed and which are supposed to actually be strings

    # A built-for binary encoding method like hex or base64 would be 'proper',
    # but doesn't really give any information to a reader. In some cases, a
    # binary blob may have embedded strings or integer values that would be
    # beneficial to quickly skim.

    # This uses latin1 encoding because it can handle arbitrary bytes, prints
    # ASCII characters and can be decoded back to the same exact byte string.
    # It's possible I missed another encoding method that matches these
    # properties across python2.7 and python3.9, but had issues with some other
    # backslash escape mechanisms parsing back to bytes.

    if config is None:
        config = default_config
    name_map = {
        item["name"]: number
        for number, item in typedef.items()
        if ("name" in item and item["name"] != "")
    }
    if not isinstance(values, dict):
        # this function should only ever be called on a message, error out if
        # it is not one. This usually means a type got swapped around
        raise EncoderException(
            "Error performing _json_safe_transform on message. Field was expected to be a message but was not: %r"
            % values
        )
    for name, value in values.items():
        if isinstance(name, int):
            name = six.ensure_text(str(name))
        alt_number = None
        base_name = name
        if "-" in name:
            base_name, alt_number = name.split("-")

        if base_name in name_map:
            field_number = name_map[base_name]
        else:
            field_number = base_name

        if field_number not in typedef or "type" not in typedef[field_number]:
            raise EncoderException(
                "Field %r not found in typedef or does not have type attribute."
                % field_number
            )

        field_type = typedef[field_number]["type"]  # type: str | TypeDefDict
        if field_type == "message":
            field_typedef = _get_typedef_for_message(typedef[field_number], config)
            if alt_number is not None:
                # if we have an alt type, then let's look that up instead
                if alt_number not in typedef[field_number].get("alt_typedefs", {}):
                    raise TypedefException(
                        (
                            "Provided alt field name/number "
                            "%s is not valid for field_number %s"
                        )
                        % (alt_number, field_number)
                    )
                field_type = typedef[field_number]["alt_typedefs"][alt_number]
                if isinstance(field_type, dict):
                    field_typedef = field_type
                    field_type = "message"

        is_list = isinstance(value, list)
        field_values = value if is_list else [value]
        for i, field_value in enumerate(field_values):
            if field_type == "bytes":
                if toBytes:
                    field_values[i] = field_value.encode("latin1")
                else:
                    field_values[i] = field_value.decode("latin1")
            elif field_type == "message":
                field_values[i] = _json_safe_transform(
                    field_value,
                    field_typedef,
                    toBytes,
                    config=config,
                )

        # convert back to single value if needed
        if not is_list:
            values[name] = field_values[0]
        else:
            values[name] = field_values
    return values


def _get_typedef_for_message(field_typedef, config):
    # type: (FieldDefDict, Config) -> TypeDefDict
    assert field_typedef["type"] == "message"
    if "message_typedef" in field_typedef:
        return field_typedef["message_typedef"]
    elif field_typedef.get("message_type_name"):
        if field_typedef["message_type_name"] not in config.known_types:
            raise TypedefException(
                "Got 'message_type_name' not in known_messages: %s"
                % field_typedef["message_type_name"]
            )
        return config.known_types[field_typedef["message_type_name"]]
    else:
        raise TypedefException(
            "Got 'message' type without typedef or type name: %s" % field_typedef
        )


def _sort_output(value, typedef, config=None):
    # type: (Message, TypeDefDict, Optional[Config]) -> Message
    # Sort output by the field number in the typedef. Helps with readability in
    # a JSON dump
    output_dict = collections.OrderedDict()  # type: Message
    if config is None:
        config = default_config

    # Make a list of all the field names we have, aggregate together the alt fields as well
    field_names = {}  # type: Dict[str, List[tuple[str, str | None]]]
    for field_name in value.keys():
        if isinstance(field_name, int):
            field_name = six.ensure_text(str(field_name))
        if "-" in field_name:
            field_name_base, alt_number = field_name.split("-")
        else:
            field_name_base = field_name
            alt_number = None
        field_names.setdefault(field_name_base, []).append((field_name, alt_number))

    for field_number, field_def in sorted(typedef.items(), key=lambda t: int(t[0])):
        field_number = six.ensure_text(str(field_number))
        seen_field_names = field_names.get(field_number, [])

        # Try getting matching fields by name as well
        if field_def.get("name", "") != "":
            field_name = field_def["name"]
            seen_field_names.extend(field_names.get(field_name, []))

        for field_name, alt_number in seen_field_names:
            field_type = field_def["type"]
            field_message_typedef = None
            if field_type == "message":
                field_message_typedef = _get_typedef_for_message(field_def, config)

            if alt_number is not None:
                if alt_number not in field_def["alt_typedefs"]:
                    raise TypedefException(
                        (
                            "Provided alt field name/number "
                            "%s is not valid for field_number %s"
                        )
                        % (alt_number, field_number)
                    )
                alt_field_type = field_def["alt_typedefs"][alt_number]
                if isinstance(alt_field_type, dict):
                    field_message_typedef = alt_field_type
                    field_type = "message"
                else:
                    field_type = alt_field_type

            if field_type == "message":
                if field_message_typedef is None:
                    raise TypedefException(
                        'Message does not have an associated typedef: "%s"' % field_name
                    )
                field_value = value.get(field_name)
                if isinstance(field_value, list):
                    output_dict[field_name] = []
                    for field_value_item in field_value:
                        if not isinstance(field_value_item, dict):
                            raise TypedefException(
                                'Message values must be a dictionary type. Field name: "%s"'
                                % field_name
                            )
                        output_dict[field_name].append(
                            _sort_output(field_value_item, field_message_typedef)
                        )
                else:
                    if not isinstance(field_value, dict):
                        raise TypedefException(
                            'Message values must be a dictionary type. Field name: "%s"'
                            % field_name
                        )
                    output_dict[field_name] = _sort_output(
                        field_value, field_message_typedef
                    )
            else:
                output_dict[field_name] = value[field_name]

    return output_dict


def sort_typedef(typedef):
    # type: (TypeDefDict) -> TypeDefDict
    """Apply special sorting rules to the type definition to improve readability.

    Sorts the fields of a type definition so that important fields such as the
    'type' or 'name' are at the top and don't get buried beneath longer fields
    like 'message_typedef'. This will also sort the keys of the
    'message_typedef' based on the field number.
    Args:
        typedef - dictionary representing a Blackboxprotobuf type definition
    Returns:
        A new OrderedDict object containing the contents of the typedef
        argument sorted for readability.
    """

    # Sort output by field number and sub_keys so name then type is first

    TYPEDEF_KEY_ORDER = [
        "name",
        "type",
        "message_type_name",
        "example_value_ignored",
        "field_order",
        "seen_repeated",
        "message_typedef",
        "alt_typedefs",
    ]
    output_dict = collections.OrderedDict()

    for field_number, field_def in sorted(
        typedef.items(), key=lambda t: int(t[0])
    ):  # Sort by type number
        output_field_def = collections.OrderedDict()
        for key, value in sorted(
            field_def.items(), key=lambda t: (TYPEDEF_KEY_ORDER.index(t[0]), t[1])
        ):  # sort by special keys, then value
            if key == "message_typedef":
                output_field_def[key] = sort_typedef(value)  # type: ignore
            else:
                output_field_def[key] = value  # type: ignore

        output_dict[field_number] = output_field_def
    if six.PY3 and typing.TYPE_CHECKING:
        return typing.cast(
            TypeDefDict, output_dict
        )  # Cast because typing doesn't like the ordered dict
    return output_dict


def _annotate_typedef(typedef, message):
    # type: (TypeDefDict, Message) -> None
    # Add values from message into the typedef so it's easier to figure out
    # which field when you're editing manually

    for field_number, field_def in typedef.items():
        field_value = None
        field_name = six.ensure_text(str(field_number))
        if field_name not in message and field_def.get("name", "") != "":
            field_name = field_def["name"]

        if field_name in message:
            field_value = message[field_name]

            if field_def["type"] == "message":
                if isinstance(field_value, list):
                    for value in field_value:
                        _annotate_typedef(field_def["message_typedef"], value)
                else:
                    _annotate_typedef(field_def["message_typedef"], field_value)
            else:
                field_def["example_value_ignored"] = field_value

        # Add a blank name field if the field doesn't have one, so it's easier
        # to add
        if "name" not in field_def:
            field_def["name"] = six.u("")


def _strip_typedef_annotations(typedef):
    # type: (TypeDefDict) -> None
    # Remove example values placed by _annotate_typedef
    for _, field_def in typedef.items():
        if "example_value_ignored" in field_def:
            del field_def["example_value_ignored"]
        if "message_typedef" in field_def:
            _strip_typedef_annotations(field_def["message_typedef"])



from burp import IBurpExtender
from burp import IContextMenuFactory, IContextMenuInvocation
from java.io import PrintWriter
from javax.swing import JMenuItem
import subprocess
from java.lang import System
import os
import commands
from burp import IMessageEditorTabFactory
from burp import IMessageEditorTab
from burp import IScannerInsertionPointProvider
from burp import IScannerInsertionPoint
from burp import IIntruderPayloadProcessor
import traceback

import grpc_utils
from burp_utils_with_dep.burp_grpc_decodetab import ProtoDecodeTab
from burp_utils_with_dep.burp_grpc_insertionpoint import GrpcInsertionPoint


class BurpExtender(IBurpExtender, IContextMenuFactory, IMessageEditorTabFactory, IScannerInsertionPointProvider, IIntruderPayloadProcessor):

    def registerExtenderCallbacks(self, callbacks):
        self.helpers = callbacks.getHelpers()
        callbacks.setExtensionName('gRPC-Web Coder Old Extension (with dependency)')
        callbacks.registerContextMenuFactory(self)
        self.stdout = PrintWriter(callbacks.getStdout(), True)
        self.stderr = PrintWriter(callbacks.getStderr(), True)
        callbacks.registerMessageEditorTabFactory(self)
        callbacks.registerScannerInsertionPointProvider(self)
        callbacks.registerIntruderPayloadProcessor(self)
        self._callbacks = callbacks

    def getInsertionPoints(self, baseRequestResponse):
        request = baseRequestResponse.getRequest()
        requestInfo = self.helpers.analyzeRequest(request)

        # check if gRPC request
        headers = requestInfo.getHeaders()
        isGrpc = False
        for header in headers:
            if header.lower().startswith("content-type: application/grpc-web-text"):
                isGrpc = True
                break

        if not isGrpc:
            return None

        body = request[requestInfo.getBodyOffset():]
        decoded = self.decode_grpc_payload(body)

        try:
            # Use the new find_insertion_points function
            found_points = grpc_utils.find_insertion_points(decoded)

            # Convert the found points into Burp insertion points
            insertionPoints = []
            for point in found_points:
                insertionPoints.append(
                    GrpcInsertionPoint(
                        self,
                        request,
                        requestInfo.getBodyOffset(),
                        decoded,
                        point['name'],  # Using the name from the found point
                        field_path=point['path']  # Using the path from the found point
                    )
                )

            return insertionPoints

        except:
            print("Error creating insertion points")
            print(traceback.format_exc())
            return None

    def getProcessorName(self):
        return "gRPC Payload Processor"

    def processPayload(self, currentPayload, originalPayload, baseValue):
        """Implement IIntruderPayloadProcessor"""
        # Encode the payload for gRPC
        try:
            payload = self.helpers.bytesToString(currentPayload)
            encoded = grpc_utils.get_encoded_payload_grpc_web_text(payload)
            return self.helpers.stringToBytes(encoded)
        except:
            return currentPayload

    def decode_grpc_payload(self, payload):
        """decode gRPC payload"""
        try:
            payload_str = self.helpers.bytesToString(payload)
            return grpc_utils.get_decoded_payload_grpc_web_text(payload_str)
        except:
            print("Error decoding gRPC payload")
            print(traceback.format_exc())
            print(str(payload))
            return payload

    def createNewInstance(self, controller, editable):
        
        return ProtoDecodeTab(self, controller, editable, self.helpers)

    def createMenuItems(self, invocation):

        context = invocation.getInvocationContext()

        selected_payload = self.get_selected_text(invocation)
        payload_index = self.get_index_of_selected_text(invocation)
        if context == IContextMenuInvocation.CONTEXT_MESSAGE_EDITOR_REQUEST \
                or context == IContextMenuInvocation.CONTEXT_MESSAGE_VIEWER_REQUEST \
                and selected_payload is not None:

            label_encode_grpc_web_text = 'Encode application/grpc-web-text'
            label_decode_grpc_web_text = 'Decode application/grpc-web-text'

            menu_item_encoding_grpc_web_text = JMenuItem(label_encode_grpc_web_text, actionPerformed=self.encode_payload_grpc_web_text)
            menu_item_encoding_grpc_web_text.putClientProperty('grpc_selected_payload', selected_payload)
            menu_item_encoding_grpc_web_text.putClientProperty('grpc_selected_payload_index', payload_index)
            menu_item_encoding_grpc_web_text.putClientProperty('grpc_encoding_invocation', invocation)

            menu_item_decoding_grpc_web_text = JMenuItem(label_decode_grpc_web_text, actionPerformed=self.decode_payload_grpc_web_text)
            menu_item_decoding_grpc_web_text.putClientProperty('grpc_selected_payload', selected_payload)
            menu_item_decoding_grpc_web_text.putClientProperty('grpc_selected_payload_index', payload_index)
            menu_item_decoding_grpc_web_text.putClientProperty('grpc_decoding_invocation', invocation)

            # label_encode_grpc_web = 'Encode application/grpc-web+proto'
            # label_decode_grpc_web = 'Decode application/grpc-web+proto'
            #
            # menu_item_encoding_grpc_web = JMenuItem(label_encode_grpc_web, actionPerformed=self.encode_payload_grpc_web)
            # menu_item_encoding_grpc_web.putClientProperty('grpc_selected_payload', selected_payload)
            # menu_item_encoding_grpc_web.putClientProperty('grpc_selected_payload_index', payload_index)
            # menu_item_encoding_grpc_web.putClientProperty('grpc_encoding_invocation', invocation)
            #
            # menu_item_decoding_grpc_web = JMenuItem(label_decode_grpc_web, actionPerformed=self.decode_payload_grpc_web)
            # menu_item_decoding_grpc_web.putClientProperty('grpc_selected_payload', selected_payload)
            # menu_item_decoding_grpc_web.putClientProperty('grpc_selected_payload_index', payload_index)
            # menu_item_decoding_grpc_web.putClientProperty('grpc_decoding_invocation', invocation)

            label_big_string_chunker = 'Chunk Big String'
            menu_item_big_string_chunker = JMenuItem(label_big_string_chunker, actionPerformed=self.chunk_big_string)
            menu_item_big_string_chunker.putClientProperty('big_string_selected_payload', selected_payload)
            menu_item_big_string_chunker.putClientProperty('big_string_selected_payload_index', payload_index)
            menu_item_big_string_chunker.putClientProperty('big_string_invocation', invocation)

            label_un_chunk_chunked_string = 'Un-Chunk Chunked String'
            menu_item_un_chunk_chunked_string = JMenuItem(
                label_un_chunk_chunked_string,
                actionPerformed=self.un_chunk_chunked_string
            )
            menu_item_un_chunk_chunked_string.putClientProperty('chunked_string_selected_payload', selected_payload)
            menu_item_un_chunk_chunked_string.putClientProperty('chunked_string_selected_payload_index', payload_index)
            menu_item_un_chunk_chunked_string.putClientProperty('chunked_string_invocation', invocation)

            return [
                menu_item_decoding_grpc_web_text,
                menu_item_encoding_grpc_web_text,
                # menu_item_decoding_grpc_web,
                # menu_item_encoding_grpc_web,
                menu_item_big_string_chunker,
                menu_item_un_chunk_chunked_string,
            ]

    def un_chunk_chunked_string(self, event):
        """
        Un-Chunk chunked string (remove ['"','\n','  '])
        :param event:
        :return:
        """

        menu_item = event.getSource()
        selected_payload = menu_item.getClientProperty('chunked_string_selected_payload')
        _invocation = menu_item.getClientProperty('chunked_string_invocation')
        _index = menu_item.getClientProperty('chunked_string_selected_payload_index')

        temp_file_path = 'chunked_string.txt'

        with open(temp_file_path, "w") as file:
            file.write(selected_payload.strip())

        python_name = "python"
        if not System.getProperty('os.name').startswith("Windows"):
            status, _ = commands.getstatusoutput('which python3')
            if status != 0:
                python_name = "python"
            else:
                python_name = "python3"

        command = python_name + " big_string_chunker.py --file chunked_string.txt --un-chunk"

        output = subprocess.check_output(command, shell=True, stderr=subprocess.STDOUT)
        # output = output.decode('utf-8')
        output = output.strip()

        # Check if the file exists before attempting to remove it
        if os.path.exists(temp_file_path):
            try:
                os.remove(temp_file_path)
            except OSError as e:
                self.pprint(str(e))

        if output:
            self.update_decoded_request_grpc_web_text(_invocation, _index, output)

        return None

    def chunk_big_string(self, event):
        """
        Chunk big string into pieces of 80 characters using big_string_chunker.py
        :param event:
        :return:
        """

        menu_item = event.getSource()
        selected_payload = menu_item.getClientProperty('big_string_selected_payload')
        _invocation = menu_item.getClientProperty('big_string_invocation')
        _index = menu_item.getClientProperty('big_string_selected_payload_index')

        temp_file_path = 'big_string.txt'

        with open(temp_file_path, "w") as file:
            file.write(selected_payload.strip())

        python_name = "python"
        if not System.getProperty('os.name').startswith("Windows"):
            status, _ = commands.getstatusoutput('which python3')
            if status != 0:
                python_name = "python"
            else:
                python_name = "python3"

        command = python_name + " big_string_chunker.py --file big_string.txt --chunk"

        output = subprocess.check_output(command, shell=True, stderr=subprocess.STDOUT)
        # output = output.decode('utf-8')
        output = output.strip()

        # Check if the file exists before attempting to remove it
        if os.path.exists(temp_file_path):
            try:
                os.remove(temp_file_path)
            except OSError as e:
                self.pprint(str(e))

        if output:
            self.update_decoded_request_grpc_web_text(_invocation, _index, output)

        return None

    def get_selected_text(self, invocation):
        request_text = invocation.getSelectedMessages()[0].getRequest().tostring().decode('utf-8')

        text_index = invocation.getSelectionBounds()
        start_index = text_index[0]
        end_index = text_index[1]

        selected_text = request_text[start_index:end_index]

        return selected_text

    def get_index_of_selected_text(self, invocation):

        text_index = invocation.getSelectionBounds()
        start_index = text_index[0]
        end_index = text_index[1]

        return (start_index, end_index)

    def update_decoded_request_grpc_web_text(self, invocation, index, new_payload):
        """
        update the request with decoded value
        application/grpc-web-text Content-Type
        :param invocation:
        :param index:
        :param new_payload:
        :return:
        """
        # index is a tuple --> (start_index, end_index)

        old_request_string = invocation.getSelectedMessages()[0].getRequest().tostring()

        new_request_string = old_request_string[:index[0]] + new_payload + old_request_string[index[1] + 1:]
        new_request_bytes = self.helpers.stringToBytes(new_request_string)
        invocation.getSelectedMessages()[0].setRequest(new_request_bytes)

    def update_decoded_request_grpc_web(self, invocation, index, new_payload):
        """
        update the request with decoded value
        application/grpc-web+proto Content-Type
        :param invocation:
        :param index:
        :param new_payload:
        :return:
        """
        # index is a tuple --> (start_index, end_index)

        old_request_string = invocation.getSelectedMessages()[0].getRequest().tostring()

        new_request_string = old_request_string[:index[0]] + new_payload + old_request_string[index[1] + 1:]
        new_request_bytes = self.helpers.stringToBytes(new_request_string)
        invocation.getSelectedMessages()[0].setRequest(new_request_bytes)

    def update_encoded_request_grpc_web_text(self, invocation, index, new_payload):
        """
        update the request with encoded value
        application/grpc-web-text Content-Type
        :param invocation:
        :param index:
        :param new_payload:
        :return:
        """
        # index is a tuple --> (start_index, end_index)

        old_request_string = invocation.getSelectedMessages()[0].getRequest().tostring().decode('utf-8')

        new_request_string = old_request_string[:index[0]] + new_payload + old_request_string[index[1] + 1:]
        new_request_bytes = self.helpers.stringToBytes(new_request_string)
        invocation.getSelectedMessages()[0].setRequest(new_request_bytes)

    def update_encoded_request_grpc_web(self, invocation, index, new_payload):
        """
        update the request with encoded value
        application/grpc-web+proto Content-Type
        :param invocation:
        :param index:
        :param new_payload:
        :return:
        """
        # index is a tuple --> (start_index, end_index)

        old_request_string = invocation.getSelectedMessages()[0].getRequest().tostring().decode('utf-8')

        new_request_string = old_request_string[:index[0]] + new_payload + old_request_string[index[1] + 1:]
        new_request_bytes = self.helpers.stringToBytes(new_request_string)
        invocation.getSelectedMessages()[0].setRequest(new_request_bytes)

    def encode_payload_grpc_web_text(self, event):
        """
        Encode application/grpc-web-text Content-Type
        :param event:
        :return:
        """

        menu_item = event.getSource()
        selected_payload = menu_item.getClientProperty('grpc_selected_payload')
        _invocation = menu_item.getClientProperty('grpc_encoding_invocation')
        _index = menu_item.getClientProperty('grpc_selected_payload_index')

        temp_file_path_encoding = 'grpc_coder_output_encode.txt'

        try:
            with open(temp_file_path_encoding, "w") as temp_encoding_file:
                temp_encoding_file.write(selected_payload.strip().encode('utf-8'))
        except Exception as e:
            self.pprint(str(e))

        python_name = "python"
        if not System.getProperty('os.name').startswith("Windows"):
            status, _ = commands.getstatusoutput('which python3')
            if status != 0:
                python_name = "python"
            else:
                python_name = "python3"

        command = grpc_utils.PROTOSCOPE_PATH + " -s grpc_coder_output_encode.txt | " + python_name + " grpc_coder.py --encode"
        output = subprocess.check_output(command, shell=True, stderr=subprocess.STDOUT)
        output = output.decode('utf-8')
        output = output.strip()

        # Check if the file exists before attempting to remove it
        if os.path.exists(temp_file_path_encoding):
            try:
                os.remove(temp_file_path_encoding)
            except OSError as e:
                self.pprint(str(e))

        if output:
            self.update_encoded_request_grpc_web_text(_invocation, _index, output)

        return None

    def decode_payload_grpc_web_text(self, event):
        """
        Decode application/grpc-web-text Content-Type
        :param event:
        :return:
        """

        menu_item = event.getSource()
        selected_payload = menu_item.getClientProperty('grpc_selected_payload')
        _invocation = menu_item.getClientProperty('grpc_decoding_invocation')
        _index = menu_item.getClientProperty('grpc_selected_payload_index')

        temp_file_path = 'grpc_coder_output_decode.txt'

        with open(temp_file_path, "w") as file:
            file.write(selected_payload.strip())

        python_name = "python"
        if not System.getProperty('os.name').startswith("Windows"):
            status, _ = commands.getstatusoutput('which python3')
            if status != 0:
                python_name = "python"
            else:
                python_name = "python3"

        command = python_name + " grpc_coder.py --decode --file grpc_coder_output_decode.txt | " + grpc_utils.PROTOSCOPE_PATH

        output = subprocess.check_output(command, shell=True, stderr=subprocess.STDOUT)
        # output = output.decode('utf-8')
        output = output.strip()

        # Check if the file exists before attempting to remove it
        if os.path.exists(temp_file_path):
            try:
                os.remove(temp_file_path)
            except OSError as e:
                self.pprint(str(e))

        if output:
            self.update_decoded_request_grpc_web_text(_invocation, _index, output)

        return None

    def encode_payload_grpc_web(self, event):
        """
       Encode application/grpc-web+proto Content-Type
       :param event:
       :return:
       """

        menu_item = event.getSource()
        selected_payload = menu_item.getClientProperty('grpc_selected_payload')
        _invocation = menu_item.getClientProperty('grpc_encoding_invocation')
        _index = menu_item.getClientProperty('grpc_selected_payload_index')

        temp_file_path_encoding = 'grpc_coder_output_encode.txt'

        try:
            with open(temp_file_path_encoding, "w") as temp_encoding_file:
                temp_encoding_file.write(selected_payload.strip())
        except Exception as e:
            self.pprint(str(e))

        python_name = "python"
        if not System.getProperty('os.name').startswith("Windows"):
            status, _ = commands.getstatusoutput('which python3')
            if status != 0:
                python_name = "python"
            else:
                python_name = "python3"

        command = grpc_utils.PROTOSCOPE_PATH + " -s grpc_coder_output_encode.txt | " + python_name + " grpc_coder.py --encode --type grpc-web+proto"
        output = subprocess.check_output(command, shell=True, stderr=subprocess.STDOUT)
        output = output.strip()

        # Check if the file exists before attempting to remove it
        if os.path.exists(temp_file_path_encoding):
            try:
                os.remove(temp_file_path_encoding)
            except OSError as e:
                self.pprint(str(e))

        if output:
            self.update_encoded_request_grpc_web(_invocation, _index, output)

        return None

    def decode_payload_grpc_web(self, event):
        """
        Decode application/grpc-web+proto Content-Type
        :param event:
        :return:
        """

        menu_item = event.getSource()
        selected_payload = menu_item.getClientProperty('grpc_selected_payload')
        _invocation = menu_item.getClientProperty('grpc_decoding_invocation')
        _index = menu_item.getClientProperty('grpc_selected_payload_index')

        temp_file_path = 'grpc_coder_output_decode.txt'

        with open(temp_file_path, "wb") as file:
            file.write(selected_payload.strip())

        python_name = "python"
        if not System.getProperty('os.name').startswith("Windows"):
            status, _ = commands.getstatusoutput('which python3')
            if status != 0:
                python_name = "python"
            else:
                python_name = "python3"

        command = python_name + " grpc_coder.py --decode --file grpc_coder_output_decode.txt --type grpc-web+proto | " + grpc_utils.PROTOSCOPE_PATH

        output = subprocess.check_output(command, shell=True, stderr=subprocess.STDOUT)
        # output = output.decode('utf-8')
        output = output.strip()

        # Check if the file exists before attempting to remove it
        if os.path.exists(temp_file_path):
            try:
                os.remove(temp_file_path)
            except OSError as e:
                self.pprint(str(e))

        if output:
            self.update_decoded_request_grpc_web(_invocation, _index, output)

        return None

    def pprint(self, text):
        self.stdout.println(text)

    def print_error(self, text):
        """
        write error
        :return:
        """
        self.stderr.println(str(text))




# Copyright (c) 2018-2024 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import json
import base64
import argparse

import typing
from typing import Any, Dict, Optional, Tuple

from .lib.exceptions import BlackboxProtobufException
from .lib import api
from .lib import payloads
from .lib.pytypes import TypeDefDict, Message


def main():
    # type: () -> int
    parser = argparse.ArgumentParser(description="Decode/Encode Protobuf Messages")
    parser.add_argument(
        "-e",
        "--encode",
        action="store_true",
        help="Switch to encode mode. Command decodes by default",
    )
    parser.add_argument(
        "-j",
        "--json-protobuf",
        action="store_true",
        help='Use JSON objects which may contain base64\'d protobuf and typedefs instead of just raw protobuf bytes for encoding input and decoding output. The JSON may have the following keys: "message", "typedef", "payload_encoding".',
    )
    parser.add_argument(
        "--compact",
        action="store_true",
        help="Use compact/non-pretty JSON output",
    )
    parser.add_argument(
        "-pe",
        "--payload-encoding",
        action="store",
        help="Override the wrapper encoding for the payload, such as gzip or grpc",
    )
    parser.add_argument(
        "-it",
        "--input-type",
        action="store",
        help="File to read the typedef from instead of stdin",
    )
    parser.add_argument(
        "-ot",
        "--output-type",
        action="store",
        help='(Decoding) File to write the typedef to instead of stdout. This file may be a plain JSON typedef or a json object with "typedef" and "payload_encoding" fields.',
    )
    parser.add_argument(
        "-r",
        "--raw-decode",
        action="store_true",
        help="(Decoding) Output just the decoded JSON and no type information.",
    )

    args = parser.parse_args()

    message = None  # type:  str | bytes | Message | None
    typedef = None  # type: TypeDefDict | None
    payload_encoding = None  # type: str | None

    if args.input_type:
        typedef, payload_encoding = _read_input_typedef_arg(args)

    if args.payload_encoding:
        payload_encoding = args.payload_encoding

    input_data = _read_input(args)
    if args.encode:
        input_json = json.loads(input_data)
        message, typedef, payload_encoding = _read_input_json_encoding(
            args, input_json, typedef, payload_encoding
        )
    elif args.json_protobuf:
        input_json = json.loads(input_data)
        message, typedef, payload_encoding = _read_input_json_decoding(
            args, input_json, typedef, payload_encoding
        )
    else:
        message = input_data

    if payload_encoding is None:
        payload_encoding = "none"

    # Start basic, raw protobuf decoding
    if args.encode:
        if not isinstance(message, dict):
            sys.stderr.write("Error did not get a valid message to encode")
            return 1
        if typedef is None:
            sys.stderr.write("Encoding requires a valid type definition")
            return 1
        return _encode(args, message, typedef, payload_encoding)
    else:
        if isinstance(message, str):
            message = message.encode("utf-8")
        if not isinstance(message, bytes):
            sys.stderr.write("Error did not get a valid message to decode")
            return 1
        return _decode(args, message, typedef, payload_encoding)


# Reads input from the location from args
# Does not handle any JSON decoding
def _read_input(args):
    # type: (argparse.Namespace) -> str | bytes
    if args.encode or args.json_protobuf:
        # Text
        return sys.stdin.read()
    else:
        # Binary
        return sys.stdin.buffer.read()


# Writes output to the location from args
# Does not handle any JSON encoding
def _write_output(args, data):
    # type: (argparse.Namespace, str | bytes) -> None
    if isinstance(data, str):
        # Text
        sys.stdout.write(data)
    else:
        # Binary
        sys.stdout.buffer.write(data)


def _read_input_typedef_arg(args):
    # type: (argparse.Namespace) -> Tuple[TypeDefDict, Optional[str]]
    with open(args.input_type, "r") as f:
        input_json = json.load(f)
    if "typedef" in input_json:
        return input_json.get("typedef"), input_json.get("payload_encoding")
    else:
        # Return whole paylaod as typedef, no encoding
        return input_json, None


def _write_output_typedef_arg(args, typedef):
    # type: (argparse.Namespace, Dict[str, str | TypeDefDict]) -> None
    with open(args.output_type, "w") as f:
        f.write(_to_json(args, typedef))


def _to_json(args, data):
    # type: (argparse.Namespace, Dict[str, Any]) -> str
    if args.compact:
        return json.dumps(data)
    else:
        return json.dumps(data, indent=2)


def _read_input_json_encoding(args, input_json, typedef, payload_encoding):
    # type: (argparse.Namespace, Message | Dict[str, str | Message | TypeDefDict], Optional[TypeDefDict], Optional[str]) -> Tuple[Message, Optional[TypeDefDict], Optional[str]]
    if typedef is None and "typedef" not in input_json:
        sys.stderr.write(
            "Error: Did not get a typedef from --input-type or stdin. A typedef is required for encoding\n"
        )
        sys.exit(1)

    message = typing.cast(Message | None, input_json.get("message"))
    if message is None:
        # Whole input is message. We already checked to make sure we have a typedef, so we can ditch it
        return typing.cast(Message, input_json), typedef, None

    if typedef is None:
        typedef = typing.cast(TypeDefDict | None, input_json.get("typedef"))

    if payload_encoding is None:
        json_payload_encoding = input_json.get("payload_encoding")
        if isinstance(json_payload_encoding, str):
            payload_encoding = json_payload_encoding
        elif json_payload_encoding is None:
            payload_encoding = None
        else:
            sys.stderr.write(
                "Warn: Payload encoding must be a string value: %r" % payload_encoding
            )
            payload_encoding = None

    return message, typedef, payload_encoding


def _read_input_json_decoding(args, input_json, typedef, payload_encoding):
    # type: (argparse.Namespace, Dict[str, TypeDefDict | str], Optional[TypeDefDict], Optional[str]) -> Tuple[bytes, Optional[TypeDefDict], Optional[str]]
    # Return message, typedef, payload_encoding
    message = typing.cast(str | None, input_json.get("protobuf_data"))
    if message is None:
        sys.stderr.write('Error: Did not get a "protobuf_data" attribute in input JSON')
        sys.exit(1)

    if typedef is None:
        typedef = typing.cast(TypeDefDict | None, input_json.get("typedef"))

    if payload_encoding is None:
        payload_encoding = typing.cast(str | None, input_json.get("payload_encoding"))

    # base64 decode if protobuf_json when decoding
    protobuf_data = base64.b64decode(message)

    return protobuf_data, typedef, payload_encoding


def _encode(args, message, typedef, payload_encoding):
    # type: (argparse.Namespace, Message, TypeDefDict, Optional[str]) -> int
    if typedef is None:
        sys.stderr.write("Error: Cannot encode without a valid typedef")
        return 1

    if not payload_encoding:
        payload_encoding = "none"

    # Re jsonify so that bbpb can fix bytes
    message_json = json.dumps(message)

    protobuf_data = api.protobuf_from_json(message_json, typedef)

    data = payloads.encode_payload(protobuf_data, payload_encoding)

    if args.json_protobuf:
        json_out = {
            "protobuf_data": base64.b64encode(data).decode("ascii"),
            "typedef": typedef,  # Typedef is a bit redundant here
        }
        if payload_encoding != "none":
            json_out["payload_encoding"] = payload_encoding

        _write_output(args, _to_json(args, json_out))
    else:
        _write_output(args, data)
    return 0


def _decode(args, data, typedef, payload_encoding):
    # type: (argparse.Namespace, bytes, Optional[TypeDefDict], str) -> int
    if len(data) == 0:
        sys.stderr.write("Error: Input data cannot be empty\n")
        return 1

    # args.protobuf_json is already handled

    if payload_encoding:
        # Use provided payload encoding algorithm
        protobuf_data, payload_encoding = payloads.decode_payload(
            data, payload_encoding
        )
        message_json, output_typedef = api.protobuf_to_json(protobuf_data, typedef)
    else:
        # Have to guess the decoding algorithm
        decoders = payloads.find_decoders(data)

        for decode in decoders:
            try:
                protobuf_data, encoding_alg = decode(data)
            except BlackboxProtobufException:
                # The "none" algorithm should always succeed
                continue

            try:
                message_json, output_typedef = api.protobuf_to_json(
                    protobuf_data, typedef
                )
                break
            except BlackboxProtobufException as exc:
                if encoding_alg == "none":
                    raise exc

    message = json.loads(message_json)

    if args.output_type:
        output_typedef_data = {}  # type: Dict[str, TypeDefDict | str]
        output_typedef_data["typedef"] = output_typedef
        if payload_encoding != "none":
            output_typedef_data["payload_encoding"] = payload_encoding
        _write_output_typedef_arg(args, output_typedef_data)

    if args.raw_decode:
        output = message
    else:
        output = {
            "message": message,
            "typedef": output_typedef,
        }
        if payload_encoding != "none":
            output["payload_encoding"] = payload_encoding
    _write_output(args, _to_json(args, output))
    return 0


if __name__ == "__main__":
    sys.exit(main())



import pytest
import blackboxprotobuf


@pytest.mark.skip()
def test_wide():
    typedef = {"1": {"type": "int"}}

    message = {"1": [1] * 10000000}

    encoded = blackboxprotobuf.lib.encode_message(message, typedef)
    decoded, _ = blackboxprotobuf.lib.decode_message(encoded, typedef)


@pytest.mark.skip()
def test_deep():
    config = blackboxprotobuf.lib.config.Config()

    typedef = {
        "1": {"type": "message", "message_type_name": "test"},
        "2": {"type": "int"},
    }
    config.known_types["test"] = typedef
    target_depth = 100
    message = {}
    last_layer = message

    while target_depth:
        new_layer = {"2": 1}
        last_layer["1"] = new_layer
        last_layer = new_layer

        target_depth -= 1

    encoded = blackboxprotobuf.lib.encode_message(message, typedef, config)
    decoded, _ = blackboxprotobuf.lib.decode_message(encoded, typedef, config)


@pytest.mark.skip()
def test_large_multilayer():
    config = blackboxprotobuf.lib.config.Config()

    typedef = {
        "1": {"type": "message", "message_type_name": "test"},
        "2": {"type": "int"},
    }
    config.known_types["test"] = typedef
    target_depth = 10
    message = {}
    last_layer = message

    while target_depth:
        new_layer = {"2": [1] * 10000}
        last_layer["1"] = [new_layer] * 2
        last_layer = new_layer

        target_depth -= 1

    encoded = blackboxprotobuf.lib.encode_message(message, typedef, config)
    decoded, _ = blackboxprotobuf.lib.decode_message(encoded, typedef, config)



"""
GRPC-Scan
Extracting Methods, Services and Messages (Routes) in JS files (grpc-web)
"""

import re
from argparse import ArgumentParser
import sys
import jsbeautifier
from texttable import Texttable


def create_table(columns_list, rows_list):
    table = Texttable()

    table_list = [columns_list]
    for i in rows_list:
        table_list.append(i)
    table.add_rows(table_list)
    table_string = table.draw()
    # indented_table_string = '    ' + table_string.replace('\n', '\n    ')  # Add space before each line

    return table_string


def beautify_js_content(content):
    try:
        beautified = jsbeautifier.beautify(content)

    except Exception as e:
        print('An error occurred in beautifying Javascript code: ' + str(e))
        print('Enter valid javascript code. Do not copy js code '
              'from browser dev tools directly! try to download it directly!')
        print('If you are still getting this error, try to beautify Javascript code online and then use this tool!')
        exit(1)

    return beautified


def extract_endpoints(content):
    pattern = r'MethodDescriptor\("(\/.*?)"'
    compiled_pattern = re.compile(pattern)
    matched_items = compiled_pattern.findall(content)
    matched_items = list(matched_items)
    print('Found Endpoints:')
    if matched_items:
        for m in matched_items:
            print("  " + m)

    print()


def extract_messages(content):
    pattern = r'proto\.(.*)\.prototype\.set(.*).*=.*function\(.*\).*{\s*.*set(.*)\(.*?,(.*?),'
    compiled_pattern = re.compile(pattern)
    matched_items = compiled_pattern.findall(content)
    matched_items = list(matched_items)

    message_list = {}

    print('Found Messages:')
    if matched_items:
        for m in matched_items:

            if m[0].strip() not in message_list:
                message_list[m[0]] = []
            if m[1].strip() not in message_list[m[0].strip()]:
                # add proto field *name* 1, add proto field *type* 2, add proto field *number* 3
                temp_list = [m[1].strip(), m[2].strip(), m[3].strip()]
                message_list[m[0]].append(temp_list)

        for m2 in message_list.keys():
            print()
            print(f'{m2}:')
            print(create_table(columns_list=['Field Name', 'Field Type', 'Field Number'], rows_list=message_list[m2]))

    print()


def read_file(file):
    try:
        with open(file, 'r', encoding='utf-8') as file:
            return file.read()

    except Exception as e:
        print('Error occurred on opening file: ' + str(e))
        exit(1)


def read_standard_input():
    return sys.stdin.read()


def print_parser_help(prog):
    help_msg = f"""python3 {prog} [INPUT]
    
    Input Arguments:
      --file      file name of js file
      --stdin     get input from standard input
    Help:
      --help      print help message
"""

    print(help_msg)


if __name__ == "__main__":
    parser = ArgumentParser(usage='python3 %(prog)s [INPUT]',
                            allow_abbrev=False, add_help=False)

    parser.add_argument('--help', action='store_true', default=False)
    parser.add_argument('--file')
    parser.add_argument('--stdin', action='store_true', default=False)

    args, unknown = parser.parse_known_args()

    if (args.help is True) or (args.file is None):
        if args.stdin is not True:
            print_parser_help(prog=parser.prog)
            exit(0)

    js_content = ""

    if args.file is not None:
        js_content = read_file(args.file)
    else:
        js_content = read_standard_input()

    js_content = beautify_js_content(js_content)

    extract_endpoints(js_content)
    extract_messages(js_content)



from burp import IBurpExtender
from burp import IContextMenuFactory, IContextMenuInvocation
from java.io import PrintWriter
from javax.swing import JMenuItem
from java.lang import System
from burp import IMessageEditorTabFactory
from burp import IMessageEditorTab
from burp import IScannerInsertionPointProvider
from burp import IScannerInsertionPoint
from burp import IIntruderPayloadProcessor
from array import array
import traceback

import grpc_utils


class GrpcInsertionPoint(IScannerInsertionPoint):
    INS_EXTENSION_PROVIDED = 65

    def __init__(self, extender, baseRequest, offset, decodedData, insertionPointName, field_path=None):
        self._extender = extender
        self._baseRequest = baseRequest
        self._offset = offset
        self._originalData = decodedData
        self._insertionPointName = insertionPointName
        self._field_path = field_path or []  # Track the path to this field in nested structure
        self._fullMessage = decodedData

        try:
            print("\nParsing:", decodedData)
            self._baseValue = grpc_utils.extract_value_from_path(self._field_path, decodedData)

        except:
            print("Error parsing gRPC structure in insertion point")
            print(traceback.format_exc())

    def buildRequest(self, payload):
        try:
            if isinstance(payload, array):
                payload = "".join(map(chr, payload))
            elif isinstance(payload, bytes):
                payload = self._extender.helpers.bytesToString(payload)

            print("Building request with payload:", payload)
            print("Current field path:", self._field_path)

            new_message = grpc_utils.replace_value_at_path(self._fullMessage, self._field_path, payload)
            print("Modified message:", new_message)

            encodedPayload = grpc_utils.get_encoded_payload_grpc_web_text(new_message)
            print("Decoded: ", grpc_utils.get_decoded_payload_grpc_web_text(encodedPayload))
            # Validate encoded payload
            if not encodedPayload:
                print("Warning: Empty encoded payload")
                return self._baseRequest

            # Construct the new request
            prefix = self._baseRequest[:self._offset]
            encoded_bytes = self._extender.helpers.stringToBytes(encodedPayload)

            # Build the complete request
            result = prefix + encoded_bytes

            # Add any remaining data after the body if needed
            if self._offset + len(encoded_bytes) < len(self._baseRequest):
                suffix = self._baseRequest[self._offset + len(encoded_bytes):]
                result += suffix

            print("Final request length:", len(result))
            return result

        except Exception as e:
            print("Error building request:", str(e))
            print(traceback.format_exc())
            return self._baseRequest

    def getPayloadOffsets(self, payload):
        print("getPayloadOffsets called with payload:", payload)
        try:
            encoded_base = grpc_utils.get_encoded_payload_grpc_web_text(self._fullMessage)
            base_bytes = self._extender.helpers.stringToBytes(encoded_base)
            start = self._offset
            end = start + len(base_bytes)
            print("Calculated offsets:", [start, end])
            return [start, end]
        except:
            print("Error calculating offsets")
            print(traceback.format_exc())
            return [self._offset, self._offset + len(payload)]

    def getInsertionPointType(self):
        print("getInsertionPointType called, returning:", self.INS_EXTENSION_PROVIDED)
        return INS_EXTENSION_PROVIDED

    def getInsertionPointName(self):
        print("getInsertionPointName called, returning:", self._insertionPointName)
        return self._insertionPointName

    def getBaseValue(self):
        print("getBaseValue called with base value:", self._baseValue)
        return self._baseValue if self._baseValue is not None else ""


# Copyright (c) 2018-2024 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import six
from blackboxprotobuf.lib.exceptions import (
    BlackboxProtobufException,
    EncoderException,
    TypedefException,
)

if six.PY3:
    import typing

    if typing.TYPE_CHECKING:
        from typing import Dict, Optional, Any, List, TypedDict, Tuple
        from blackboxprotobuf.lib.config import Config
        from .pytypes import TypeDefDict, FieldDefDict


class TypeDef(object):
    def __init__(self):
        # type: (TypeDef) -> None
        self._fields = {}  # type: Dict[str, FieldDef]
        self._field_names = {}  # type: Dict[str, str]

    @staticmethod
    def from_dict(typedef_dict):
        # type: (TypeDefDict) -> TypeDef
        typedef = TypeDef()
        for field_id, fielddef_dict in typedef_dict.items():
            fielddef = FieldDef.from_dict(fielddef_dict, field_id)
            typedef._fields[field_id] = fielddef
            if fielddef._name:
                typedef._field_names[fielddef._name] = field_id
        return typedef

    def to_dict(self):
        # type: (TypeDef) -> TypeDefDict
        return {field: fielddef.to_dict() for field, fielddef in self._fields.items()}

    def make_mutable(self):
        # type: (TypeDef) -> MutableTypeDef
        mutable = MutableTypeDef()
        # Copy fields but don't deep copy
        # They should get immutable copies of FieldDef
        mutable._fields = self._fields.copy()

        return mutable

    def lookup_fielddef(self, key):
        # type: (TypeDef, str) -> Optional[Tuple[str, FieldDef]]
        """Look up a field definition by number (w/ alt typedef) or name"""
        # We don't care what the alt number is now
        field_name = key.split("-", 1)[0]

        field_id = self._field_names.get(field_name, field_name)

        if field_id in self._fields:
            return field_id, self._fields[field_id]

        return None

    def lookup_fielddef_number(self, field_id):
        # type: (TypeDef, str) -> Optional[Tuple[str, FieldDef]]
        if field_id in self._fields:
            return field_id, self._fields[field_id]
        return None


class MutableTypeDef(TypeDef):
    def set_fielddef(self, field_number, fielddef):
        # type: (MutableTypeDef, str, FieldDef) -> None
        self._fields[field_number] = fielddef
        if fielddef._name:
            self._field_names[fielddef._name] = field_number


class FieldDef(object):
    def __init__(self, field_id):
        # type: (FieldDef, str) -> None
        self._name = None  # type: Optional[str]
        self._field_id = field_id  # type: str
        self._message_type_name = None  # type: Optional[str]
        # Normal type will be 0, alts starting at 1
        # Each field will have either the type or a typedef
        # We don't allow message_type_name in alt_typedefs, we'll use it for any "message" entry instead
        self._types = {}  # type: Dict[str, str | TypeDef]
        self._example_value = None  # type: Any
        self._seen_repeated = False  # type: bool
        self._field_order = None  # type: Optional[List[str]]

    @staticmethod
    def from_dict(fielddef_dict, field_id):
        # type: (FieldDefDict, str) -> FieldDef
        fielddef = FieldDef(field_id)

        if "name" in fielddef_dict:
            fielddef._name = fielddef_dict["name"]

        if "type" in fielddef_dict:
            field_type = fielddef_dict["type"]

            if field_type == "message" and "message_typedef" in fielddef_dict:
                fielddef._types["0"] = TypeDef.from_dict(
                    fielddef_dict["message_typedef"]
                )
            else:
                fielddef._types["0"] = field_type

        if "message_type_name" in fielddef_dict:
            # We could lookup the typedef, but better to wait to resolve
            fielddef._message_type_name = fielddef_dict["message_type_name"]

        if "alt_typedefs" in fielddef_dict:
            for alt_number, alt_typedef in fielddef_dict["alt_typedefs"].items():
                if isinstance(alt_typedef, dict):
                    fielddef._types[alt_number] = TypeDef.from_dict(alt_typedef)
                else:
                    fielddef._types[alt_number] = alt_typedef

        if "example_value_ignored" in fielddef_dict:
            fielddef._example_value = fielddef_dict["example_value_ignored"]

        if "seen_repeated" in fielddef_dict:
            fielddef._seen_repeated = fielddef_dict["seen_repeated"]

        if "field_order" in fielddef_dict:
            fielddef._field_order = fielddef_dict["field_order"]

        return fielddef

    def to_dict(self):
        # type: (FieldDef) -> FieldDefDict
        fielddef_dict = {}  # type: FieldDefDict
        if self._name:
            fielddef_dict["name"] = self._name
        if self._message_type_name:
            fielddef_dict["message_type_name"] = self._message_type_name
        if self._example_value:
            fielddef_dict["example_value_ignored"] = self._example_value
        if self._seen_repeated:
            fielddef_dict["seen_repeated"] = self._seen_repeated
        if self._field_order:
            fielddef_dict["field_order"] = self._field_order

        field_type = self._types.get("0")
        if isinstance(field_type, TypeDef):
            field_typedef = field_type.to_dict()
            field_type = "message"
            fielddef_dict["message_typedef"] = field_typedef

        if field_type:
            fielddef_dict["type"] = field_type

        if field_type and len(self._types) > 1:
            fielddef_dict["alt_typedefs"] = {
                alt_num: (
                    alt_type.to_dict() if isinstance(alt_type, TypeDef) else alt_type
                )
                for alt_num, alt_type in self._types.items()
                if alt_num != "0"
            }

        return fielddef_dict

    def make_mutable(self):
        # type: (FieldDef) -> MutableFieldDef
        mutable = MutableFieldDef(self._field_id)
        mutable._name = self._name
        mutable._message_type_name = self._message_type_name
        mutable._types = self._types.copy()
        mutable._example_value = self._example_value
        mutable._seen_repeated = self._seen_repeated
        mutable._field_order = self._field_order
        return mutable

    def lookup_field_type(self, key, config, field_path):
        # type: (FieldDef, str, Config, List[str]) -> Optional[str | TypeDef]
        if "-" in key:
            alt_type_id = key.split("-", 1)[1]
        else:
            alt_type_id = "0"

        return self.lookup_field_type_number(alt_type_id, config, field_path)

    # Lookup fieled type by just alt type number
    def lookup_field_type_number(self, alt_type_id, config, field_path):
        # type: (FieldDef, str, Config, List[str]) -> Optional[str | TypeDef]

        if alt_type_id not in self._types:
            return None

        field_type = self._types[alt_type_id]
        if field_type == "message":
            # We have to look up the message type name
            return self.resolve_message_type_name(config, field_path)

        return field_type

    @property
    def name(self):
        # type: (FieldDef) -> str
        if self._name:
            return self._name
        return self._field_id

    def field_key(self, alt_field_id):
        # type: (FieldDef, str) -> str
        if alt_field_id == "0":
            return self.name
        else:
            return self.name + "-" + alt_field_id

    def next_alt_type_id(self):
        # type: (FieldDef) -> str
        existing_ids = [int(alt_type_id) for alt_type_id in self._types.keys()]
        if len(existing_ids) == 0:
            return "0"
        else:
            return six.ensure_text(str(max(existing_ids) + 1))

    @property
    def field_order(self):
        # type: (FieldDef) -> Optional[List[str]]
        # someone could mutate the list returned, but that's not a high
        # priority
        return self._field_order

    @property
    def seen_repeated(self):
        # type: (FieldDef) -> bool
        return self._seen_repeated

    def resolve_message_type_name(self, config, field_path):
        # type: (FieldDef, Config, List[str]) -> TypeDef
        if self._message_type_name not in config.known_types:
            raise TypedefException(
                "Message name '%s' has not been defined in known types"
                % self._message_type_name,
                field_path,
            )
        return TypeDef.from_dict(config.known_types[self._message_type_name])

    def resolve_types(self, config, field_path):
        # type: (FieldDef, Config, List[str]) -> Dict[str, str | TypeDef]
        field_types = self._types.copy()
        if field_types.get("0") == "message":
            field_types["0"] = self.resolve_message_type_name(config, field_path)
        return field_types


class MutableFieldDef(FieldDef):
    def set_field_order(self, field_order):
        # type: (MutableFieldDef, List[str]) -> None
        self._field_order = field_order

    def mark_repeated(self):
        # type: (MutableFieldDef) -> None
        self._seen_repeated = True

    def set_type(self, alt_type_id, field_type):
        # type: (MutableFieldDef, str, str | TypeDef) -> None
        self._types[alt_type_id] = field_type

    def set_types(self, types):
        # type: (MutableFieldDef, Dict[str, str | TypeDef]) -> None
        self._types = types

    def add_type(self, field_type):
        # type: (MutableFieldDef, str | TypeDef) -> str
        alt_type_id = self.next_alt_type_id()
        self.set_type(alt_type_id, field_type)
        return alt_type_id



# grpc_utils.py
import subprocess
import os

import traceback
# Add at the top of the file:
import sys
reload(sys)
sys.setdefaultencoding('utf-8')

PROTOSCOPE_PATH = "protoscope"


def extract_value_from_path(field_path, message):
    """Extract the value at the specified field path"""
    if not field_path:
        return message

    lines = message.split('\n')
    path_stack = []

    for line in lines:
        stripped = line.lstrip()
        indent = len(line) - len(stripped)

        while path_stack and indent <= path_stack[-1][0]:
            path_stack.pop()

        if ': {' in stripped:
            field_num = stripped.split(':', 1)[0].strip()
            value = stripped.split('{', 1)[1].rstrip('}').strip()

            path_stack.append((indent, field_num))
            current_path = [p[1] for p in path_stack]

            if current_path == field_path:
                if value.startswith('"') and value.endswith('"'):
                    return value[1:-1]
                return value

    return None


def replace_value_at_path(message, field_path, new_value):
    """Replace value at the specified field path (escaping included)"""

    def escape_value(val):
        val = str(val)
        val = val.replace('\\', '\\\\').replace('"', '\\"')
        val = val.replace('\n', '\\n').replace('\r', '\\r')
        return '"{}"'.format(val)

    if not field_path:
        return message

    lines = message.split('\n')
    result = []
    path_stack = []

    for line in lines:
        stripped = line.lstrip()
        indent = len(line) - len(stripped)

        while path_stack and indent <= path_stack[-1][0]:
            path_stack.pop()

        if ': {' in stripped:
            field_num = stripped.split(':', 1)[0].strip()
            path_stack.append((indent, field_num))

            current_path = [p[1] for p in path_stack]
            if current_path == field_path:
                escaped = escape_value(new_value)
                new_line = ' ' * indent + field_num + ': {' + escaped + '}'
                result.append(new_line)
                continue

        result.append(line)

    return '\n'.join(result)


def find_insertion_points(decoded_message):
    """Find all possible insertion points in a decoded protobuf message"""
    insertion_points = []

    def parse_message(message, current_path=[]):
        """Recursively parse protobuf message to find all fields"""
        lines = message.split('\n')
        path_stack = []
        current_indent = 0

        for line in lines:
            stripped = line.strip()
            if not stripped:
                continue

            indent = len(line) - len(stripped)

            # Adjust path stack based on indentation
            while path_stack and indent <= path_stack[-1][0]:
                path_stack.pop()

            if ': {' in stripped:
                field_num = stripped.split(':', 1)[0].strip()
                value = stripped.split('{', 1)[1].rstrip('}').strip()

                path_stack.append((indent, field_num))
                new_path = [p[1] for p in path_stack]

                if value == '':
                    continue
                # Add the current field as an insertion point
                insertion_points.append({
                    'path': new_path,
                    'name': "gRPC field {}".format('.'.join(new_path)),
                    'value': value
                })

                # If the value contains nested structure, parse it
                if value.count('{') > value.count('"'):
                    parse_message(value, new_path)

    parse_message(decoded_message)
    return insertion_points


def get_decoded_payload_grpc_web_text(payload):
    temp_file_path = 'grpc_coder_output_decode.txt'
    temp_proto_path = 'proto_output.txt'

    try:
        # If payload is bytes, keep it as-is for base64
        if isinstance(payload, bytes):
            raw_payload = payload
        else:
            # If string/unicode, encode to ascii, ignoring problematic chars
            raw_payload = payload.encode('ascii', 'ignore')

        # Write raw bytes
        with open(temp_file_path, 'wb') as file:
            file.write(raw_payload)

        python_name = "python"
        if not os.name.startswith("Windows"):
            try:
                status = subprocess.check_output('which python3', shell=True)
                python_name = "python3"
            except subprocess.CalledProcessError:
                python_name = "python"

        command = [python_name, "grpc_coder.py", "--decode", "--file", temp_file_path]
        decoded = subprocess.check_output(command, shell=False)

        with open(temp_proto_path, 'wb') as f:
            f.write(decoded)

        try:
            protoscope_command = [PROTOSCOPE_PATH, temp_proto_path]
            protoscope_output = subprocess.check_output(protoscope_command, shell=False)
            return protoscope_output
        except subprocess.CalledProcessError as e:
            return decoded

    except:
        return payload
    finally:
        for temp_file in [temp_file_path, temp_proto_path]:
            if os.path.exists(temp_file):
                try:
                    os.remove(temp_file)
                except:
                    pass


def get_encoded_payload_grpc_web_text(payload):
    temp_file_path_encoding = 'grpc_coder_output_encode.txt'
    temp_proto_path = 'proto_output.txt'

    try:
        # If payload is bytes, keep it as-is for base64
        if isinstance(payload, bytes):
            raw_payload = payload
        else:
            # If string/unicode, encode to ascii, ignoring problematic chars
            raw_payload = payload.encode('ascii', 'ignore')

        # Write raw bytes
        with open(temp_file_path_encoding, 'wb') as file:
            file.write(raw_payload)

        python_name = "python"
        if not os.name.startswith("Windows"):
            try:
                status = subprocess.check_output('which python3', shell=True)
                python_name = "python3"
            except subprocess.CalledProcessError:
                python_name = "python"

        try:
            protoscope_command = [PROTOSCOPE_PATH, "-s", temp_file_path_encoding]
            with open(temp_proto_path, 'wb') as f:
                subprocess.check_call(protoscope_command, stdout=f, shell=False)

            encode_command = [python_name, "grpc_coder.py", "--encode", "--file", temp_proto_path]
            output = subprocess.check_output(encode_command, shell=False)
            return output.strip()

        except subprocess.CalledProcessError:
            encode_command = [python_name, "grpc_coder.py", "--encode", "--file", temp_file_path_encoding]
            output = subprocess.check_output(encode_command, shell=False)
            return output.strip()

    except:
        return payload
    finally:
        for temp_file in [temp_file_path_encoding, temp_proto_path]:
            if os.path.exists(temp_file):
                try:
                    os.remove(temp_file)
                except:
                    pass


from burp import IBurpExtender, IMessageEditorTabFactory, ITab
from java.io import PrintWriter
from javax.swing import JPanel, JCheckBox, JLabel, BoxLayout
from java.awt.event import ActionListener
from burp_utils.burp_grpc_web_editor_tab import GrpcWebExtensionEditorTab


class BurpExtender(IBurpExtender, IMessageEditorTabFactory, ITab, ActionListener):
    def registerExtenderCallbacks(self, callbacks):
        self._callbacks = callbacks
        self._helpers = callbacks.getHelpers()

        # Name of extension
        callbacks.setExtensionName("gRPC-Web Pentest Suite")

        # Set stdout and stderr
        self.stdout = PrintWriter(callbacks.getStdout(), True)
        self.stderr = PrintWriter(callbacks.getStderr(), True)

        # Register TabFactory
        callbacks.registerMessageEditorTabFactory(self)

        # Create and register a new UI tab
        self._panel = JPanel()
        self._panel.setLayout(BoxLayout(self._panel, BoxLayout.Y_AXIS))

        features_label = JLabel("------ Features ------")

        self._enable_extension_msg_editor_tab_check_box = JCheckBox("Enable Extension Message Editor gRPC-Web Tab by "
                                                                    "Default on "
                                                                    "all Requests")
        self._enable_extension_msg_editor_tab_check_box.addActionListener(self)
        self._enable_extension_msg_editor_tab_check_box.setSelected(True)
        self._detect_encode_decode_format_from_ct_header_check_box = JCheckBox("Detect Encode/Decode format from "
                                                                               "Content-Type header (with value of "
                                                                               "application/grpc-web-text or "
                                                                               "application/grpc-web+proto)")
        self._detect_encode_decode_format_from_ct_header_check_box.addActionListener(self)
        self._detect_encode_decode_format_from_ct_header_check_box.setSelected(True)
        self._detect_encode_decode_format_from_x_grpc_header_check_box = JCheckBox("Detect Encode/Decode format from "
                                                                                   "x-grpc-content-type header (with "
                                                                                   "value of "
                                                                                   "application/grpc-web-text or "
                                                                                   "application/grpc-web+proto)")

        self._detect_encode_decode_format_from_x_grpc_header_check_box.addActionListener(self)
        self._detect_encode_decode_format_from_x_grpc_header_check_box.setEnabled(False)
        content_type_note_label = JLabel("------ If you enable following checkbox, extension tries to decode/encode "
                                         "the body on every request ------")
        self._enable_application_grpc_web_text_decode_encode_by_default_check_box = JCheckBox("Enable "
                                                                                              "application/grpc-web"
                                                                                              "-text Decode/Encode by "
                                                                                              "default on all requests")
        self._enable_application_grpc_web_text_decode_encode_by_default_check_box.addActionListener(self)

        self._enable_application_grpc_web_proto_decode_encode_by_default_check_box = JCheckBox("Enable "
                                                                                              "application/grpc-web"
                                                                                              "+proto Decode/Encode by "
                                                                                              "default on all requests")
        self._enable_application_grpc_web_proto_decode_encode_by_default_check_box.addActionListener(self)

        # Add components to panel
        self._panel.add(features_label)
        self._panel.add(self._enable_extension_msg_editor_tab_check_box)
        self._panel.add(self._detect_encode_decode_format_from_ct_header_check_box)
        self._panel.add(self._detect_encode_decode_format_from_x_grpc_header_check_box)
        self._panel.add(content_type_note_label)
        self._panel.add(self._enable_application_grpc_web_text_decode_encode_by_default_check_box)
        self._panel.add(self._enable_application_grpc_web_proto_decode_encode_by_default_check_box)

        # Register the new tab
        callbacks.addSuiteTab(self)

        # Handle checkbox clicks
    def actionPerformed(self, event):
        source = event.getSource()
        if source == self._enable_extension_msg_editor_tab_check_box:
            pass

        elif source == self._detect_encode_decode_format_from_ct_header_check_box:
            if self._detect_encode_decode_format_from_ct_header_check_box.isSelected():
                self._detect_encode_decode_format_from_x_grpc_header_check_box.setEnabled(False)
            else:
                self._detect_encode_decode_format_from_x_grpc_header_check_box.setEnabled(True)

        elif source == self._detect_encode_decode_format_from_x_grpc_header_check_box:
            if self._detect_encode_decode_format_from_x_grpc_header_check_box.isSelected():
                self._detect_encode_decode_format_from_ct_header_check_box.setEnabled(False)
            else:
                self._detect_encode_decode_format_from_ct_header_check_box.setEnabled(True)

        elif source == self._enable_application_grpc_web_text_decode_encode_by_default_check_box:
            if self._enable_application_grpc_web_text_decode_encode_by_default_check_box.isSelected():
                self._detect_encode_decode_format_from_x_grpc_header_check_box.setSelected(False)
                self._detect_encode_decode_format_from_x_grpc_header_check_box.setEnabled(False)

                self._detect_encode_decode_format_from_ct_header_check_box.setSelected(False)
                self._detect_encode_decode_format_from_ct_header_check_box.setEnabled(False)

                self._enable_extension_msg_editor_tab_check_box.setSelected(True)
                self._enable_extension_msg_editor_tab_check_box.setEnabled(False)

                self._enable_application_grpc_web_proto_decode_encode_by_default_check_box.setSelected(False)
                self._enable_application_grpc_web_proto_decode_encode_by_default_check_box.setEnabled(False)

            else:
                self._detect_encode_decode_format_from_x_grpc_header_check_box.setEnabled(True)
                self._detect_encode_decode_format_from_ct_header_check_box.setEnabled(True)

                self._enable_extension_msg_editor_tab_check_box.setSelected(True)
                self._enable_extension_msg_editor_tab_check_box.setEnabled(True)

                self._enable_application_grpc_web_proto_decode_encode_by_default_check_box.setEnabled(True)

        elif source == self._enable_application_grpc_web_proto_decode_encode_by_default_check_box:
            if self._enable_application_grpc_web_proto_decode_encode_by_default_check_box.isSelected():
                self._detect_encode_decode_format_from_x_grpc_header_check_box.setSelected(False)
                self._detect_encode_decode_format_from_x_grpc_header_check_box.setEnabled(False)

                self._detect_encode_decode_format_from_ct_header_check_box.setSelected(False)
                self._detect_encode_decode_format_from_ct_header_check_box.setEnabled(False)

                self._enable_extension_msg_editor_tab_check_box.setSelected(True)
                self._enable_extension_msg_editor_tab_check_box.setEnabled(False)

                self._enable_application_grpc_web_text_decode_encode_by_default_check_box.setSelected(False)
                self._enable_application_grpc_web_text_decode_encode_by_default_check_box.setEnabled(False)
            else:
                self._detect_encode_decode_format_from_x_grpc_header_check_box.setEnabled(True)
                self._detect_encode_decode_format_from_ct_header_check_box.setEnabled(True)

                self._enable_extension_msg_editor_tab_check_box.setSelected(True)
                self._enable_extension_msg_editor_tab_check_box.setEnabled(True)

                self._enable_application_grpc_web_text_decode_encode_by_default_check_box.setEnabled(True)

    def getTabCaption(self):
        return "gRPC-Web Pentest Suite"

    def getUiComponent(self):
        return self._panel

    def createNewInstance(self, controller, editable):
        return GrpcWebExtensionEditorTab(self, controller, editable)

    def print_output(self, text):
        self.stdout.println(text)

    def print_error(self, text):
        self.stderr.println(text)



# Copyright (c) 2018-2024 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from hypothesis import given, example, note, assume
import hypothesis.strategies as st
import strategies
import pytest
import six

from google.protobuf.internal import wire_format, encoder, decoder

from blackboxprotobuf.lib.types import varint
from blackboxprotobuf.lib.exceptions import EncoderException, DecoderException


# Test that for any given bytes, we don't alter them when decoding as a varint
@given(x=st.binary(max_size=10))
@example(x=b"\x80\x01")
@example(x=b"\x80\x81")
@example(x=b"\x81\x80\x80\x80\x80\x80\x80\x80\x01\x00")
@example(x=b"\x80\x80\x80\x80\x80\x80\x80\x80\x81\x00")
@example(x=b"\x80\x80\x80\x80\x80\x80\x80\x80\x81\x80")
def test_varint_idem_uvarint(x):
    try:
        decoded, pos = varint.decode_uvarint(x, 0)
    except DecoderException:
        assume(True)
        return

    encoded = varint.encode_uvarint(decoded)
    assert encoded == x[:pos]


# Test that for any given bytes, we don't alter them when decoding as a varint
@given(x=st.binary(min_size=10, max_size=10))
@example(x=b"\x80\x01")
@example(x=b"\x80\x81")
@example(x=b"\x81\x80\x80\x80\x80\x80\x80\x80\x01\x00")
@example(x=b"\x80\x80\x80\x80\x80\x80\x80\x80\x81\x00")
@example(x=b"\x80\x80\x80\x80\x80\x80\x80\x80\x81\x80")
@example(x=b"\x8d\x9b\xb0\xcc\xcf\xdc\xea\xf4\xf9\x02")
def test_varint_idem_varint(x):
    try:
        decoded, pos = varint.decode_varint(x, 0)
    except DecoderException:
        assume(True)
        return
    encoded = varint.encode_varint(decoded)
    assert encoded == x[:pos]


# Test that for any given bytes, we don't alter them when decoding as a varint
@given(x=st.binary(max_size=10))
@example(x=b"\x80\x01")
@example(x=b"\x80\x81")
@example(x=b"\x81\x80\x80\x80\x80\x80\x80\x80\x01\x00")
@example(x=b"\x80\x80\x80\x80\x80\x80\x80\x80\x81\x00")
@example(
    x=b"\x80\x80\x80\x80\x80\x80\x80\x80\x81\x80"
)  # I think this is overflowing and getting truncated on decode
def test_varint_idem_svarint(x):
    try:
        decoded, pos = varint.decode_svarint(x, 0)
    except DecoderException:
        assume(True)
        return
    encoded = varint.encode_svarint(decoded)
    assert encoded == x[:pos]


# Inverse checks. Ensure a value encoded by bbp decodes to the same value
@given(x=strategies.input_map["uint"])
@example(x=18446744073709551615)
def test_uvarint_inverse(x):
    encoded = varint.encode_uvarint(x)
    decoded, pos = varint.decode_uvarint(encoded, 0)
    assert pos == len(encoded)
    assert decoded == x


@given(x=strategies.input_map["int"])
@example(x=-1143843382522404608)
@example(x=-1)
@example(x=8784740448578833805)
def test_varint_inverse(x):
    encoded = varint.encode_varint(x)
    decoded, pos = varint.decode_varint(encoded, 0)
    assert pos == len(encoded)
    assert decoded == x


@given(x=strategies.input_map["sint"])
def test_svarint_inverse(x):
    encoded = varint.encode_svarint(x)
    decoded, pos = varint.decode_svarint(encoded, 0)
    assert pos == len(encoded)
    assert decoded == x


@given(x=st.integers(min_value=varint.MAX_UVARINT + 1))
def test_bounds_varints(x):
    with pytest.raises(EncoderException):
        varint.encode_uvarint(x)

    with pytest.raises(EncoderException):
        varint.encode_uvarint(-x)

    with pytest.raises(EncoderException):
        varint.encode_varint(x)

    with pytest.raises(EncoderException):
        varint.encode_varint(-x)

    with pytest.raises(EncoderException):
        varint.encode_svarint(x)

    with pytest.raises(EncoderException):
        varint.encode_svarint(-x)


def _gen_append_bytearray(arr):
    def _append_bytearray(x):
        if isinstance(x, (str, int)):
            arr.append(x)
        elif isinstance(x, bytes):
            arr.extend(x)
        else:
            raise EncoderException("Unknown type returned by protobuf library")

    return _append_bytearray


# Test our varint functions against google
@given(x=strategies.input_map["uint"])
def test_uvarint_encode(x):
    encoded_google = bytearray()
    encoder._EncodeVarint(_gen_append_bytearray(encoded_google), x)
    encoded_bbpb = varint.encode_uvarint(x)
    assert encoded_google == encoded_bbpb


@given(x=strategies.input_map["uint"])
def test_uvarint_decode(x):
    buf = bytearray()
    encoder._EncodeVarint(_gen_append_bytearray(buf), x)

    if six.PY2:
        buf = str(buf)
    decoded_google, _ = decoder._DecodeVarint(buf, 0)
    decoded_bbpb, _ = varint.decode_uvarint(buf, 0)
    assert decoded_google == decoded_bbpb


@given(x=strategies.input_map["int"])
def test_varint_encode(x):
    encoded_google = bytearray()
    encoder._EncodeSignedVarint(_gen_append_bytearray(encoded_google), x)
    encoded_bbpb = varint.encode_varint(x)
    assert encoded_google == encoded_bbpb


@given(x=strategies.input_map["int"])
def test_varint_decode(x):
    buf = bytearray()
    encoder._EncodeSignedVarint(_gen_append_bytearray(buf), x)

    if six.PY2:
        buf = bytes(buf)
    decoded_google, _ = decoder._DecodeSignedVarint(buf, 0)
    decoded_bbpb, _ = varint.decode_varint(buf, 0)
    assert decoded_google == decoded_bbpb


@given(x=strategies.input_map["sint"])
def test_svarint_encode(x):
    encoded_google = bytearray()
    encoder._EncodeSignedVarint(
        _gen_append_bytearray(encoded_google), wire_format.ZigZagEncode(x)
    )
    encoded_bbpb = varint.encode_svarint(x)
    assert encoded_google == encoded_bbpb


@given(x=strategies.input_map["sint"])
@example(x=-1)
def test_svarint_decode(x):
    buf = bytearray()
    encoder._EncodeSignedVarint(_gen_append_bytearray(buf), wire_format.ZigZagEncode(x))

    if six.PY2:
        buf = bytes(buf)
    decoded_google_uint, _ = decoder._DecodeVarint(buf, 0)
    decoded_google = wire_format.ZigZagDecode(decoded_google_uint)
    decoded_bbpb, _ = varint.decode_svarint(buf, 0)

    assert decoded_google == decoded_bbpb



# Copyright (c) 2018-2024 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import six
import binascii
import hypothesis.strategies as st
import blackboxprotobuf
from blackboxprotobuf.lib.types import varint
from blackboxprotobuf.lib.types import type_maps

from hypothesis import settings
from hypothesis import database
from datetime import timedelta

settings.register_profile(
    "quick",
    max_examples=100,
    database=database.ExampleDatabase(".hypothesis-db"),
    print_blob=True,
)
settings.register_profile(
    "extended",
    max_examples=1000,
    database=database.ExampleDatabase(".hypothesis-db"),
    print_blob=True,
)
settings.register_profile(
    "superextended",
    max_examples=10000,
    database=database.ExampleDatabase(".hypothesis-db"),
    print_blob=True,
)
settings.load_profile("quick")


@st.composite
def message_typedef_gen(draw, max_depth=3, anon=False, types=None, named_fields=True):
    output = {}
    field_numbers = draw(
        st.lists(st.integers(min_value=1, max_value=2000).map(str), min_size=1)
    )
    # pre-generate names so we can be sure they're unique
    field_names = draw(
        st.lists(
            st.from_regex(blackboxprotobuf.NAME_REGEX),
            min_size=len(field_numbers),
            max_size=len(field_numbers),
            unique_by=lambda x: x.lower(),
        )
    )
    if types is None:
        message_types = [
            field_type
            for field_type in type_maps.WIRETYPES.keys()
            if field_type in input_map and input_map[field_type] is not None
        ]
    else:
        message_types = types

    for field_number, field_name in zip(field_numbers, field_names):
        field_number = six.ensure_text(str(field_number))
        if max_depth == 0 and "message" in message_types:
            message_types.remove("message")
        field_type = draw(st.sampled_from(message_types))
        output[field_number] = {}
        output[field_number]["type"] = field_type

        if field_type.startswith("packed"):
            output[field_number]["seen_repeated"] = True
        elif anon and field_type in [
            "message",
            "string",
            "bytes",
        ]:  # repeated fields can't be tested easily with anonymous types due to the type guessing
            output[field_number]["seen_repeated"] = False
        else:
            output[field_number]["seen_repeated"] = draw(st.booleans())

        if field_type == "message":
            output[field_number]["message_typedef"] = draw(
                message_typedef_gen(
                    max_depth=max_depth - 1,
                    anon=anon,
                    types=types,
                    named_fields=named_fields,
                )
            )

        # decide whether to give it a name
        if named_fields and not anon and draw(st.booleans()):
            output[field_number]["name"] = six.ensure_text(field_name)

    return output


@st.composite
def gen_message_data(draw, type_def):
    output = {}
    for number, field in type_def.items():
        if "name" in field and field["name"] != "":
            field_label = field["name"]
        else:
            field_label = six.ensure_text(str(number))

        field_type = field["type"]
        if field_type == "message":
            strat = gen_message_data(field["message_typedef"])
        else:
            strat = input_map[field["type"]]

        if field.get("seen_repeated", False) and not field_type.startswith("packed"):
            output[field_label] = draw(st.lists(strat, min_size=2))
        else:
            output[field_label] = draw(strat)

    return output


@st.composite
# if anon is True, typedef will only have "default" types that it will decode
# to without a typedef
def gen_message(draw, anon=False, named_fields=True):
    allowed_types = None
    if anon:
        allowed_types = list(
            filter(lambda x: x is not None, type_maps.WIRE_TYPE_DEFAULTS.values())
        )
        # add length delim wiretypes
        allowed_types += ["message", "string", "bytes"]
    type_def = draw(
        message_typedef_gen(anon=anon, types=allowed_types, named_fields=named_fields)
    )
    message = draw(gen_message_data(type_def))
    return type_def, message


# Map types to generators
input_map = {
    "fixed32": st.integers(min_value=0, max_value=(2**32) - 1),
    "sfixed32": st.integers(min_value=-(2**31), max_value=(2**31) - 1),
    "fixed64": st.integers(min_value=0, max_value=(2**64) - 1),
    "sfixed64": st.integers(min_value=-(2**63), max_value=(2**63) - 1),
    "float": st.floats(width=32, allow_nan=False),
    "double": st.floats(width=64, allow_nan=False),
    "uint": st.integers(min_value=varint.MIN_UVARINT, max_value=varint.MAX_UVARINT),
    "int": st.integers(min_value=varint.MIN_SVARINT, max_value=varint.MAX_SVARINT),
    "sint": st.integers(min_value=varint.MIN_SVARINT, max_value=varint.MAX_SVARINT),
    "bytes": st.binary(),
    "string": st.text(),
    #'bytes_hex':  st.binary().map(binascii.hexlify),
    "message": gen_message(),
}
input_map.update(
    {
        "packed_uint": st.lists(input_map["uint"], min_size=1),
        "packed_int": st.lists(input_map["int"], min_size=1),
        "packed_sint": st.lists(input_map["sint"], min_size=1),
        "packed_fixed32": st.lists(input_map["fixed32"], min_size=1),
        "packed_sfixed32": st.lists(input_map["sfixed32"], min_size=1),
        "packed_float": st.lists(input_map["float"], min_size=1),
        "packed_fixed64": st.lists(input_map["fixed64"], min_size=1),
        "packed_sfixed64": st.lists(input_map["sfixed64"], min_size=1),
        "packed_double": st.lists(input_map["double"], min_size=1),
        "packed_bytes": st.lists(input_map["bytes"], min_size=1),
        "packed_string": st.lists(input_map["string"], min_size=1),
        "packed_bytes": st.lists(input_map["bytes"], min_size=1),
        #'packed_bytes_hex': st.lists(input_map['bytes_hex'], min_size=1),
    }
)



# Copyright (c) 2018-2024 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import Test_pb2
import zlib
import struct
import asyncio
from websockets.server import serve

payload_type = "grpc"


async def handle_messages(websocket):
    async for message in websocket:
        print(f"Got message: {type(message)} {message}")
        data = message
        if payload_type == "gzip":
            data = zlib.decompress(data, wbits=31)
        elif payload_type == "grpc":
            old_data = data
            compression_byte = data[0]
            assert compression_byte == 0
            length = struct.unpack_from(">I", data[1:])[0]
            data = old_data[5:]
            assert length == len(data)
        print("Got data: %s" % data)
        message = Test_pb2.TestMessage()
        message.ParseFromString(data)
        print("Got message: %s" % data)

        message.testString += "_server"
        output = message.SerializeToString()

        if payload_type == "gzip":
            output = zlib.compress(output, level=9, wbits=31)
        elif payload_type == "grpc":
            # Fake grpc wrapper
            length = len(output)
            old_output = output
            output = bytearray()
            output.append(0x00)
            output.extend(struct.pack(">I", length))
            output.extend(old_output)

        await websocket.send(output)


async def main():
    async with serve(handle_messages, "localhost", 8000):
        await asyncio.Future()  # run forever


asyncio.run(main())



"""
Encode and Decode GRPC-Web Base64 Encoded Payload for Pentesting GRPC-Web
"""

import base64
import binascii
import sys
from argparse import ArgumentParser


def decode_b64_payload(b64_content):
    try:
        decoded = base64.b64decode(b64_content)
    except Exception as e:
        print('Error occurred while decoding b64 payload: ' + str(e))
        raise e

    return decoded


def encode_b64_payload(content_input):
    base64_encoded = base64.b64encode(content_input)
    return base64_encoded.decode('utf-8')


def convert_to_hex(content):
    try:
        hex_rep = binascii.hexlify(content)
    except Exception as e:
        print('Error occurred while converting payload to hex: ' + str(e))
        raise e

    return hex_rep


def new_method_convert_hex_to_ascii(hex_input):
    ascii_bytes = bytearray.fromhex(hex_input)
    return ascii_bytes


def split_grpc_length_prefix(hex_input):
    """
    split length prefix and payload from hex input
    :param hex_input:
    :return: length_prefix, payload
    """
    hex_input = hex_input.decode()
    length_prefix = hex_input[0:10]
    payload = hex_input[10:]

    return length_prefix, payload


def calculate_length_from_length_prefix(length_prefix):
    try:
        tmp = int(length_prefix, 16) * 2  # * 2 is bcs each byte has 2 characters
    except Exception as e:
        print('Error occurred while calculating length of payload: ' + str(e))
        raise e

    return tmp


def read_payload_based_on_length(payload, length):
    temp_str = payload[0:length]
    return temp_str


def convert_payload_hex_to_formatted_output(hex_payload):
    # convert for example 0a0d02 to \\x0a\\x0d\\x02

    temp_str = ""
    for i in range(0, len(hex_payload)):

        if i % 2 == 0:
            temp_str += r"\\x" + hex_payload[i]
        else:
            temp_str += hex_payload[i]

    return temp_str


def convert_hex_to_ascii(hex_input):
    return bytes.fromhex(hex_input)


def convert_ascii_to_b64(ascii_input):
    encoded_b64 = base64.b64encode(ascii_input)
    return encoded_b64.decode('utf-8')


def get_padded_length_of_new_payload(payload):
    length = len(payload) / 2
    length = int(length)
    tmp = format(length, 'x')

    if len(tmp) < 10:
        tmp = "0" * (10 - len(tmp)) + tmp

    return tmp


def decoder(content_input):
    """
    application/grpc-web-text decoder
    :param content_input:
    :return:
    """

    base64_decoded = decode_b64_payload(content_input)
    b64_to_hex = convert_to_hex(base64_decoded)
    payload_length_prefix, payload = split_grpc_length_prefix(b64_to_hex)
    length = calculate_length_from_length_prefix(payload_length_prefix)
    main_payload = read_payload_based_on_length(payload, length)
    # formatted_output = convert_payload_hex_to_formatted_output(main_payload)
    ascii_payload = convert_hex_to_ascii(main_payload)

    sys.stdout.buffer.write(ascii_payload)


def encoder(content_input):
    """
    application/grpc-web-text encoder
    :param content_input:
    :return:
    """

    hex_converted = convert_to_hex(content_input)
    hex_length_prefix = get_padded_length_of_new_payload(hex_converted)
    new_payload_with_length_prefix = hex_length_prefix + str(hex_converted.decode())
    ascii_result = convert_hex_to_ascii(new_payload_with_length_prefix)
    b64_result = convert_ascii_to_b64(ascii_result)
    print(b64_result)


def grpc_web_encoder(content_input):
    """
    application/grpc-web+proto encoder
    :param content_input:
    :return:
    """

    hex_converted = convert_to_hex(content_input)
    hex_length_prefix = get_padded_length_of_new_payload(hex_converted)
    new_payload_with_length_prefix = hex_length_prefix + str(hex_converted.decode())
    ascii_payload = convert_hex_to_ascii(new_payload_with_length_prefix)

    sys.stdout.buffer.write(ascii_payload)


def grpc_web_decoder(content_input):
    """
    application/grpc-web-text decoder
    :param content_input:
    :return:
    """

    base64_encoded_content = encode_b64_payload(content_input)
    decoder(base64_encoded_content)


def print_parser_help(prog):
    help_msg = """echo payload | python3 {} [--encode OR --decode]

    General Arguments:
      --encode       encode protoscope binary output to application/grpc-web-text
      --decode       decode application/grpc-web-text base64 encoded payload to protoscope format
      --type         content-type of payload [default: grpc-web-text] available types: [grpc-web-text, grpc-web+proto]
    
    Input Arguments:
    Default Input is Standard Input
      --file        to get input from a file 
    
    Help:
      --help        print help message
""".format(prog)

    print(help_msg)


def get_content_from_stdin():
    return sys.stdin.buffer.read()


def get_content_from_file(file_path):
    try:
        with open(file_path, 'rb') as file:
            return file.read()

    except Exception as e:
        print('Error Occurred in Reading Input File: ' + str(e))
        raise e


if __name__ == '__main__':
    parser = ArgumentParser(usage='echo payload | python3 %(prog)s [--encode or --decode]',
                            allow_abbrev=False, add_help=False)

    parser.add_argument('--help', action='store_true', default=False)
    parser.add_argument('--encode', action='store_true')
    parser.add_argument('--decode', action='store_true')
    parser.add_argument('--file', default=None)
    parser.add_argument('--type', default='grpc-web-text')

    args, unknown = parser.parse_known_args()

    if (args.encode is not True) and (args.decode is not True):
        print_parser_help(parser.prog)
        exit(1)

    if args.file is None:
        content = get_content_from_stdin()
    else:
        content = get_content_from_file(file_path=args.file)
    if args.decode is True:
        if args.type == 'grpc-web-text':
            decoder(content)
        elif args.type == 'grpc-web+proto':
            grpc_web_decoder(content)
        else:
            print('Invalid Content-Type: ' + args.type)
            print('Available types: grpc-web-text OR grpc-web+proto')
    else:
        if args.type == 'grpc-web-text':
            encoder(content)
        elif args.type == 'grpc-web+proto':
            grpc_web_encoder(content)
        else:
            print('Invalid Content-Type: ' + args.type)
            print('Available types: grpc-web-text OR grpc-web+proto')



"""Contains various maps for protobuf types, including encoding/decoding
functions, wiretypes and default types
"""

# Copyright (c) 2018-2024 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from blackboxprotobuf.lib.types import varint, fixed, length_delim, wiretypes

import six

if six.PY3:
    from typing import Any, Callable, Dict, Tuple

# Map a blackboxprotobuf type to specific encoder
ENCODERS = {
    "uint": varint.encode_uvarint,
    "int": varint.encode_varint,
    "sint": varint.encode_svarint,
    "fixed32": fixed.encode_fixed32,
    "sfixed32": fixed.encode_sfixed32,
    "float": fixed.encode_float,
    "fixed64": fixed.encode_fixed64,
    "sfixed64": fixed.encode_sfixed64,
    "double": fixed.encode_double,
    "bytes": length_delim.encode_bytes,
    "bytes_hex": length_delim.encode_bytes_hex,
    "string": length_delim.encode_string,
    "packed_uint": length_delim.generate_packed_encoder(varint.encode_uvarint),
    "packed_int": length_delim.generate_packed_encoder(varint.encode_varint),
    "packed_sint": length_delim.generate_packed_encoder(varint.encode_svarint),
    "packed_fixed32": length_delim.generate_packed_encoder(fixed.encode_fixed32),
    "packed_sfixed32": length_delim.generate_packed_encoder(fixed.encode_sfixed32),
    "packed_float": length_delim.generate_packed_encoder(fixed.encode_float),
    "packed_fixed64": length_delim.generate_packed_encoder(fixed.encode_fixed64),
    "packed_sfixed64": length_delim.generate_packed_encoder(fixed.encode_sfixed64),
    "packed_double": length_delim.generate_packed_encoder(fixed.encode_double),
}  # type: Dict[str, Callable[[Any], bytes]]

# Map a blackboxprotobuf type to specific decoder
DECODERS = {
    "uint": varint.decode_uvarint,
    "int": varint.decode_varint,
    "sint": varint.decode_svarint,
    "fixed32": fixed.decode_fixed32,
    "sfixed32": fixed.decode_sfixed32,
    "float": fixed.decode_float,
    "fixed64": fixed.decode_fixed64,
    "sfixed64": fixed.decode_sfixed64,
    "double": fixed.decode_double,
    "bytes": length_delim.decode_bytes,
    "bytes_hex": length_delim.decode_bytes_hex,
    "string": length_delim.decode_string,
    "packed_uint": length_delim.generate_packed_decoder(varint.decode_uvarint),
    "packed_int": length_delim.generate_packed_decoder(varint.decode_varint),
    "packed_sint": length_delim.generate_packed_decoder(varint.decode_svarint),
    "packed_fixed32": length_delim.generate_packed_decoder(fixed.decode_fixed32),
    "packed_sfixed32": length_delim.generate_packed_decoder(fixed.decode_sfixed32),
    "packed_float": length_delim.generate_packed_decoder(fixed.decode_float),
    "packed_fixed64": length_delim.generate_packed_decoder(fixed.decode_fixed64),
    "packed_sfixed64": length_delim.generate_packed_decoder(fixed.decode_sfixed64),
    "packed_double": length_delim.generate_packed_decoder(fixed.decode_double),
}  # type: Dict[str, Callable[[bytes, int], Tuple[Any, int]  ]]

WIRETYPES = {
    "uint": wiretypes.VARINT,
    "int": wiretypes.VARINT,
    "sint": wiretypes.VARINT,
    "fixed32": wiretypes.FIXED32,
    "sfixed32": wiretypes.FIXED32,
    "float": wiretypes.FIXED32,
    "fixed64": wiretypes.FIXED64,
    "sfixed64": wiretypes.FIXED64,
    "double": wiretypes.FIXED64,
    "bytes": wiretypes.LENGTH_DELIMITED,
    "bytes_hex": wiretypes.LENGTH_DELIMITED,
    "string": wiretypes.LENGTH_DELIMITED,
    "message": wiretypes.LENGTH_DELIMITED,
    "group": wiretypes.START_GROUP,
    "packed_uint": wiretypes.LENGTH_DELIMITED,
    "packed_int": wiretypes.LENGTH_DELIMITED,
    "packed_sint": wiretypes.LENGTH_DELIMITED,
    "packed_fixed32": wiretypes.LENGTH_DELIMITED,
    "packed_sfixed32": wiretypes.LENGTH_DELIMITED,
    "packed_float": wiretypes.LENGTH_DELIMITED,
    "packed_fixed64": wiretypes.LENGTH_DELIMITED,
    "packed_sfixed64": wiretypes.LENGTH_DELIMITED,
    "packed_double": wiretypes.LENGTH_DELIMITED,
}  # type: Dict[str, int]

# Default values to use when decoding each wire type
# length delimited is special and handled in the length_delim module
WIRE_TYPE_DEFAULTS = {
    wiretypes.VARINT: "int",
    wiretypes.FIXED32: "fixed32",
    wiretypes.FIXED64: "fixed64",
    wiretypes.LENGTH_DELIMITED: None,
    wiretypes.START_GROUP: None,
    wiretypes.END_GROUP: None,
}  # type: Dict[int, str | None]
